"""Generate the invariant verifiers from the intermediate representation."""
import io
import textwrap
from typing import (
    Tuple,
    Optional,
    List,
    Sequence,
    Mapping,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    assert_never,
    Identifier,
    indent_but_first_line,
    wrap_text_into_lines,
    assert_union_without_excluded,
)
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference
from aas_core_codegen.parse import tree as parse_tree, retree as parse_retree
from aas_core_codegen.typescript import (
    common as typescript_common,
    naming as typescript_naming,
    description as typescript_description,
    transpilation as typescript_transpilation,
)
from aas_core_codegen.typescript.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# region Verify


def verify(
    spec_impls: specific_implementations.SpecificImplementations,
    verification_functions: Sequence[intermediate.Verification],
) -> Optional[List[str]]:
    """Verify all the implementation snippets related to verification."""
    errors = []  # type: List[str]

    expected_keys = []  # type: List[specific_implementations.ImplementationKey]

    for func in verification_functions:
        if isinstance(func, intermediate.ImplementationSpecificVerification):
            expected_keys.append(
                specific_implementations.ImplementationKey(
                    f"Verification/{func.name}.ts"
                ),
            )

    for key in expected_keys:
        if key not in spec_impls:
            errors.append(f"The implementation snippet is missing for: {key}")

    if len(errors) == 0:
        return None

    return errors


# endregion

# region Generate


class _RegexRendererForJavaScript(parse_retree.Renderer):
    r"""
    Render the regular expressions as expected by JavaScript.

    For example, unicode characters are rendered as ``\u{...}``.
    """

    def char_to_str_and_escape_or_encode_if_necessary(
        self, node: parse_retree.Char, escaping: Mapping[str, str]
    ) -> List[Union[str, parse_tree.FormattedValue]]:
        """Convert the ``node`` to a string, and escape and/or encode appropriately."""
        if not node.explicitly_encoded:
            escaped = escaping.get(node.character, None)
            if escaped is not None:
                result: List[Union[str, parse_tree.FormattedValue]] = [escaped]
            else:
                result = [node.character]

            return result
        else:
            code = ord(node.character)
            # pylint: disable=line-too-long
            # NOTE (mristin, 2022-12-09):
            # See: https://dmitripavlutin.com/what-every-javascript-developer-should-know-about-unicode/#24-surrogate-pairs
            return [f"\\u{{{code:x}}}"]


_REGEX_RENDERER_FOR_JAVASCRIPT = _RegexRendererForJavaScript()


class _PatternVerificationTranspiler(
    parse_tree.RestrictedTransformer[Tuple[Optional[Stripped], Optional[Error]]]
):
    """Transpile a statement of a pattern verification into TypeScript."""

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def _transform_joined_str_values(
        self, values: Sequence[Union[str, parse_tree.FormattedValue]]
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        """Transform the values of a joined string to a TypeScript string literal."""
        # If we do not need interpolation, simply return the string literals
        # joined together.
        needs_interpolation = any(
            isinstance(value, parse_tree.FormattedValue) for value in values
        )
        if not needs_interpolation:
            return (
                Stripped(
                    typescript_common.string_literal(
                        "".join(value for value in values)  # type: ignore
                    )
                ),
                None,
            )

        parts = []  # type: List[str]

        for value in values:
            if isinstance(value, str):
                parts.append(
                    typescript_common.string_literal(
                        value,
                        without_enclosing=True,
                        in_backticks=True,
                    )
                )

            elif isinstance(value, parse_tree.FormattedValue):
                code, error = self.transform(value.value)
                if error is not None:
                    return None, error

                assert code is not None

                assert (
                    "\n" not in code
                ), f"New-lines are not expected in formatted values, but got: {code}"

                parts.append(f"${{{code}}}")
            else:
                assert_never(value)

        parts_joined = "".join(parts)
        return Stripped(f"`{parts_joined}`"), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_constant(
        self, node: parse_tree.Constant
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if isinstance(node.value, str):
            # NOTE (mristin, 2022-06-11):
            # We assume that all the string constants are valid regular expressions.

            regex, parse_error = parse_retree.parse(values=[node.value])
            if parse_error is not None:
                regex_line, pointer_line = parse_retree.render_pointer(
                    parse_error.cursor
                )

                return (
                    None,
                    Error(
                        node.original_node,
                        f"The string constant could not be parsed "
                        f"as a regular expression: \n"
                        f"{parse_error.message}\n"
                        f"{regex_line}\n"
                        f"{pointer_line}",
                    ),
                )

            assert regex is not None

            # NOTE (mristin, 2022-11-04):
            # Strictly speaking, this is a joined string with a single value, a string
            # literal. Thus, do not be confused by the name of the function —
            # this function treats both joined formatted values *and* string literals.
            return self._transform_joined_str_values(
                values=parse_retree.render(
                    regex=regex, renderer=_REGEX_RENDERER_FOR_JAVASCRIPT
                )
            )
        else:
            raise AssertionError(f"Unexpected {node=}")

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        return Stripped(typescript_naming.variable_name(node.identifier)), None

    @ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
    def transform_joined_str(
        self, node: parse_tree.JoinedStr
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        regex, parse_error = parse_retree.parse(values=node.values)
        if parse_error is not None:
            regex_line, pointer_line = parse_retree.render_pointer(parse_error.cursor)

            return (
                None,
                Error(
                    node.original_node,
                    f"The joined string could not be parsed "
                    f"as a regular expression: \n"
                    f"{parse_error.message}\n"
                    f"{regex_line}\n"
                    f"{pointer_line}",
                ),
            )

        assert regex is not None

        return self._transform_joined_str_values(
            values=parse_retree.render(
                regex=regex, renderer=_REGEX_RENDERER_FOR_JAVASCRIPT
            )
        )

    def transform_assignment(
        self, node: parse_tree.Assignment
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        assert isinstance(node.target, parse_tree.Name)
        variable = typescript_naming.variable_name(node.target.identifier)
        code, error = self.transform(node.value)
        if error is not None:
            return None, error
        assert code is not None

        # NOTE (mristin, 2022-11-24):
        # We assume that the variables won't change in the patterns. If this assumption
        # is broken, fix the code here by first inspecting the scope and classifying
        # the variables into constant ones and modifiable ones.
        return Stripped(f"const {variable} = {code};"), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_pattern_verification(
    verification: intermediate.PatternVerification,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the verification function that checks the regular expressions."""
    # NOTE (mristin, 2022-11-12):
    # We assume that we performed all the checks at the intermediate stage.

    construct_name = typescript_naming.function_name(
        Identifier(f"construct_{verification.name}")
    )

    blocks = []  # type: List[Stripped]

    # region Construct block

    writer = io.StringIO()
    writer.write(
        f"""\
function {construct_name}(): RegExp {{
"""
    )

    transpiler = _PatternVerificationTranspiler()

    for i, stmt in enumerate(verification.parsed.body):
        if i == len(verification.parsed.body) - 1:
            break

        code, error = transpiler.transform(stmt)
        if error is not None:
            return None, error
        assert code is not None

        writer.write(textwrap.indent(code, I))
        writer.write("\n")

    if len(verification.parsed.body) >= 2:
        writer.write("\n")

    pattern_expr, error = transpiler.transform(verification.pattern_expr)
    if error is not None:
        return None, error
    assert pattern_expr is not None

    # A pragmatic heuristics for breaking lines
    if len(pattern_expr) < 50:
        writer.write(textwrap.indent(f'return new RegExp({pattern_expr}, "u");', I))
    else:
        writer.write(
            textwrap.indent(
                f"""\
return new RegExp(
{I}{indent_but_first_line(pattern_expr, I)},
{I}"u"
);""",
                I,
            )
        )

    writer.write("\n}")

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    # region Initialize the regex

    regex_name = typescript_naming.constant_name(
        Identifier(f"regexp_{verification.name}")
    )

    blocks.append(Stripped(f"const {regex_name} = {construct_name}();"))

    # endregion

    # region Define the verification function

    assert len(verification.arguments) == 1
    assert isinstance(
        verification.arguments[0].type_annotation, intermediate.PrimitiveTypeAnnotation
    )
    # noinspection PyUnresolvedReferences
    assert (
        verification.arguments[0].type_annotation.a_type
        == intermediate.PrimitiveType.STR
    )

    arg_name = typescript_naming.argument_name(verification.arguments[0].name)

    function_name = typescript_naming.function_name(verification.name)

    writer = io.StringIO()

    if verification.description is not None:
        (
            comment,
            comment_errors,
        ) = typescript_description.generate_documentation_comment_for_signature(
            description=verification.description,
            context=typescript_description.Context(
                module=typescript_common.VERIFICATION_MODULE, cls_or_enum=None
            ),
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                f"Failed to generate the documentation comment for {verification.name!r}",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    writer.write(
        f"""\
export function {function_name}({arg_name}: string): boolean {{
{I}return {regex_name}.test({arg_name});
}}"""
    )

    blocks.append(Stripped(writer.getvalue()))

    # endregion

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    return Stripped(writer.getvalue()), None


class _TranspilableVerificationTranspiler(typescript_transpilation.Transpiler):
    """Transpile the body of a :py:class:`TranspilableVerification`."""

    # fmt: off
    @require(
        lambda environment, verification:
        all(
            environment.find(arg.name) is not None
            for arg in verification.arguments
        ),
        "All arguments defined in the environment"
    )
    # fmt: on
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
        verification: intermediate.TranspilableVerification,
    ) -> None:
        """Initialize with the given values."""
        typescript_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

        self._argument_name_set = frozenset(arg.name for arg in verification.arguments)

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(typescript_naming.variable_name(node.identifier)), None

        if node.identifier in self._argument_name_set:
            return Stripped(typescript_naming.argument_name(node.identifier)), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_name = typescript_naming.constant_name(node.identifier)
            return Stripped(f"AasConstants.{constant_name}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(typescript_naming.function_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return (
                Stripped(f"AasTypes.{typescript_naming.enum_name(node.identifier)}"),
                None,
            )

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to TypeScript. We could not find it neither in the constants, nor in "
            f"verification functions, nor as an enumeration. "
            f"If you expect this name to be transpilable, please contact "
            f"the developers.",
        )


def _transpile_transpilable_verification(
    verification: intermediate.TranspilableVerification,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Transpile a verification function."""
    # fmt: off
    type_inference, error = (
        intermediate_type_inference.infer_for_verification(
            verification=verification,
            base_environment=environment
        )
    )
    # fmt: on

    if error is not None:
        return None, error

    assert type_inference is not None

    transpiler = _TranspilableVerificationTranspiler(
        type_map=type_inference.type_map,
        environment=type_inference.environment_with_args,
        symbol_table=symbol_table,
        verification=verification,
    )

    body = []  # type: List[Stripped]
    for node in verification.parsed.body:
        stmt, error = transpiler.transform(node)
        if error is not None:
            return None, Error(
                verification.parsed.node,
                f"Failed to transpile the verification function {verification.name!r}",
                [error],
            )

        assert stmt is not None
        body.append(stmt)

    writer = io.StringIO()

    if verification.description is not None:
        (
            comment,
            comment_errors,
        ) = typescript_description.generate_documentation_comment_for_signature(
            description=verification.description,
            context=typescript_description.Context(
                module=typescript_common.VERIFICATION_MODULE,
                cls_or_enum=None,
            ),
        )
        if comment_errors is not None:
            return None, Error(
                verification.description.parsed.node,
                f"Failed to generate the comment for verification function {verification.name!r}",
                comment_errors,
            )

        assert comment is not None

        writer.write(comment)
        writer.write("\n")

    function_name = typescript_naming.function_name(verification.name)

    if verification.returns is None:
        return_type = "void"
    else:
        return_type = typescript_common.generate_type(
            type_annotation=verification.returns, types_module=Identifier("aas_types")
        )

    arg_defs = []  # type: List[Stripped]
    for arg in verification.arguments:
        arg_type = typescript_common.generate_type(
            arg.type_annotation, types_module=Identifier("AasTypes")
        )
        arg_name = typescript_naming.argument_name(arg.name)
        arg_defs.append(Stripped(f"{arg_name}: {arg_type}"))

    if len(arg_defs) == 0:
        writer.write(
            f"""\
export function {function_name}(): {return_type}:"""
        )
    else:
        writer.write(
            f"""\
export function {function_name}(
"""
        )

        for i, arg_def in enumerate(arg_defs):
            if i > 0:
                writer.write(",\n")
            writer.write(textwrap.indent(arg_def, I))

        writer.write("\n")
        writer.write(
            f"""\
): {return_type} {{"""
        )

    if len(body) == 0:
        writer.write("// Intentionally empty.")
    else:
        for stmt in body:
            writer.write("\n")
            writer.write(textwrap.indent(stmt, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


class _InvariantTranspiler(typescript_transpilation.Transpiler):
    def __init__(
        self,
        type_map: Mapping[
            parse_tree.Node, intermediate_type_inference.TypeAnnotationUnion
        ],
        environment: intermediate_type_inference.Environment,
        symbol_table: intermediate.SymbolTable,
    ) -> None:
        """Initialize with the given values."""
        typescript_transpilation.Transpiler.__init__(
            self, type_map=type_map, environment=environment
        )

        self._symbol_table = symbol_table

    def transform_name(
        self, node: parse_tree.Name
    ) -> Tuple[Optional[Stripped], Optional[Error]]:
        if node.identifier in self._variable_name_set:
            return Stripped(typescript_naming.variable_name(node.identifier)), None

        if node.identifier == "self":
            # The ``that`` refers to the argument of the verification function.
            return Stripped("that"), None

        if node.identifier in self._symbol_table.constants_by_name:
            constant_name = typescript_naming.constant_name(node.identifier)
            return Stripped(f"AasConstants.{constant_name}"), None

        if node.identifier in self._symbol_table.verification_functions_by_name:
            return Stripped(typescript_naming.function_name(node.identifier)), None

        our_type = self._symbol_table.find_our_type(name=node.identifier)
        if isinstance(our_type, intermediate.Enumeration):
            return (
                Stripped(f"AasTypes.{typescript_naming.enum_name(node.identifier)}"),
                None,
            )

        return None, Error(
            node.original_node,
            f"We can not determine how to transpile the name {node.identifier!r} "
            f"to TypeScript. We could not find it neither in the local variables, "
            f"nor in the global constants, nor in verification functions, "
            f"nor as an enumeration. If you expect this name to be transpilable, "
            f"please contact the developers.",
        )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _transpile_invariant(
    invariant: intermediate.Invariant,
    symbol_table: intermediate.SymbolTable,
    environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Translate the invariant from the meta-model into TypeScript code."""
    # fmt: off
    type_map, inference_error = (
        intermediate_type_inference.infer_for_invariant(
            invariant=invariant,
            environment=environment
        )
    )
    # fmt: on

    if inference_error is not None:
        return None, inference_error

    assert type_map is not None

    transpiler = _InvariantTranspiler(
        type_map=type_map,
        environment=environment,
        symbol_table=symbol_table,
    )

    expr, error = transpiler.transform(invariant.parsed.body)
    if error is not None:
        return None, error

    assert expr is not None

    writer = io.StringIO()
    if len(expr) > 50 or "\n" in expr:
        writer.write("if (!(\n")
        writer.write(textwrap.indent(expr, I))
        writer.write("\n)) {\n")
    else:
        no_parenthesis_type_in_this_context = (
            parse_tree.Index,
            parse_tree.Name,
            parse_tree.Member,
            parse_tree.MethodCall,
            parse_tree.FunctionCall,
        )

        if isinstance(invariant.parsed.body, no_parenthesis_type_in_this_context):
            not_expr = f"!{expr}"
        else:
            not_expr = f"!({expr})"

        writer.write(f"if ({not_expr}) {{\n")

    writer.write(f"{I}yield new VerificationError(\n")

    # NOTE (mristin, 2022-11-12):
    # We need to wrap the description in multiple literals as a single long
    # string literal is often too much for the readability.
    invariant_description_lines = wrap_text_into_lines(invariant.description)

    for i, literal in enumerate(invariant_description_lines):
        if i < len(invariant_description_lines) - 1:
            writer.write(f"{II}{typescript_common.string_literal(literal)} +\n")
        else:
            writer.write(f"{II}{typescript_common.string_literal(literal)}\n")
            writer.write(f"{I})")

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


OurTypeExceptEnumeration = Union[
    intermediate.ConstrainedPrimitive,
    intermediate.AbstractClass,
    intermediate.ConcreteClass,
]
assert_union_without_excluded(
    original_union=intermediate.OurType,
    subset_union=OurTypeExceptEnumeration,
    excluded=[intermediate.Enumeration],
)


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_verify_property_snippet(
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """
    Generate the snippet to transform a property to verification errors.

    Return an empty string if there is nothing to be verified for the given property.
    """
    # NOTE (mristin, 2022-11-12):
    # Instead of writing here a complex but general solution with unrolling we choose
    # to provide a simple, but limited, solution. First, the meta-model is quite
    # limited itself at the moment, so the complexity of the general solution is not
    # warranted. Second, we hope that there will be fewer bugs in the simple solution
    # which is particularly important at this early adoption stage.
    #
    # We anticipate that in the future we will indeed need a general and complex
    # solution. Here are just some thoughts on how to approach it:
    # * Leave the pattern matching to produce more readable code for simple cases,
    # * Unroll only in case of composite types and optional composite types.

    type_anno = (
        prop.type_annotation
        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation)
        else prop.type_annotation.value
    )

    if isinstance(type_anno, intermediate.OptionalTypeAnnotation):
        return None, Error(
            prop.parsed.node,
            "We currently implemented verification based on a very limited "
            "pattern matching due to code simplicity. We did not handle "
            "the case of nested optional values. Please contact "
            "the developers if you need this functionality.",
        )
    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        if isinstance(type_anno.items, intermediate.OptionalTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of optional values. Please contact "
                "the developers if you need this functionality.",
            )
        elif isinstance(type_anno.items, intermediate.ListTypeAnnotation):
            return None, Error(
                prop.parsed.node,
                "We currently implemented verification based on a very limited "
                "pattern matching due to code simplicity. We did not handle "
                "the case of lists of lists. Please contact "
                "the developers if you need this functionality.",
            )
        else:
            pass
    else:
        pass

    stmts = []  # type: List[Stripped]

    prop_name = typescript_naming.property_name(prop.name)
    prop_name_literal = typescript_common.string_literal(prop_name)

    primitive_type = intermediate.try_primitive_type(type_anno)
    if primitive_type is not None and primitive_type is intermediate.PrimitiveType.INT:
        stmts.append(
            Stripped(
                f"""\
if (!Number.isInteger(that.{prop_name})) {{
{I}const error = new VerificationError(
{II}"Expected an integer, but got a floating-point number"
{I});
{I}error.path.prepend(
{II}new PropertySegment(
{III}that,
{III}{prop_name_literal}
{II})
{I});
{I}yield error;
}}"""
            )
        )

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        # There is nothing more that we check for primitive types explicitly. The values
        # of the primitive properties are checked at the level of class invariants.
        return Stripped(""), None
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        if isinstance(type_anno.our_type, intermediate.Enumeration):
            # We rely on TypeScript compiler to check for valid enumerations, so we do not check
            # the enumerations on our side.
            return Stripped(""), None

        elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
            function_name = typescript_naming.function_name(
                Identifier(f"verify_{type_anno.our_type.name}")
            )

            # fmt: off
            for_error_of_verify = (
                f"for (const error of {function_name}(that.{prop_name}))"
            )
            # fmt: on

            # Heuristic to break the lines, very rudimentary
            if len(for_error_of_verify) > 70:
                for_error_of_verify = f"""\
for (const error of {function_name}(
{II}that.{prop_name})
)"""

            stmts.append(
                Stripped(
                    f"""\
{for_error_of_verify} {{
{I}error.path.prepend(
{II}new PropertySegment(
{III}that,
{III}{prop_name_literal}
{II})
{I});
{I}yield error;
}}"""
                )
            )

        elif isinstance(
            type_anno.our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            for_error_of_this_transform = f"for (const error of this.transformWithContext(that.{prop_name}, context))"
            # Heuristic to break the lines, very rudimentary
            if len(for_error_of_this_transform) > 70:
                for_error_of_this_transform = f"""\
for (const error of this.transformWithContext(
{II}that.{prop_name}, context)
)"""

            stmts.append(
                Stripped(
                    f"""\
{for_error_of_this_transform} {{
{I}error.path.prepend(
{II}new PropertySegment(
{III}that,
{III}{prop_name_literal}
{II})
{I});
{I}yield error;
}}"""
                )
            )
        else:
            assert_never(type_anno.our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        assert isinstance(type_anno.items, intermediate.OurTypeAnnotation), (
            "We chose to implement only a very limited pattern matching; "
            "see the note above in the code."
        )

        index_var = typescript_naming.variable_name(Identifier(f"{prop.name}_index"))

        stmts.append(
            Stripped(
                f"""\
let {index_var} = 0;
for (const item of that.{prop_name}) {{
{I}for (const error of this.transformWithContext(item, context)) {{
{II}error.path.prepend(
{III}new IndexSegment(
{IIII}that.{prop_name},
{IIII}{index_var}
{III})
{II});
{II}error.path.prepend(
{III}new PropertySegment(
{IIII}that,
{IIII}{prop_name_literal}
{III})
{II});
{II}yield error;
{I}}}
{I}{index_var}++;
}}"""
            )
        )

    else:
        assert_never(type_anno)

    verify_block = Stripped("\n".join(stmts))

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        return (
            Stripped(
                f"""\
if (that.{prop_name} !== null) {{
{I}{indent_but_first_line(verify_block, I)}
}}"""
            ),
            None,
        )

    return verify_block, None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_transform_for_class(
    cls: intermediate.ConcreteClass,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the transform method to errors for the given concrete class."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(our_type=cls),
    )

    # region Generate the non-recursive part verifying the invariants

    for invariant in cls.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"Failed to transpile the invariant of the class {cls.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    # endregion

    # region Split properties in two, constrained primitives and others

    # We split the properties in two, those of constrained primitives and others.
    # We do this in a separate step instead of duplicating the code to avoid bugs
    # related to DRY on the splitting logic.

    properties_constrained_primitives = []  # type: List[intermediate.Property]
    properties_non_constrained_primitives = []  # type: List[intermediate.Property]
    for prop in cls.properties:
        prop_type = intermediate.beneath_optional(prop.type_annotation)
        if isinstance(prop_type, intermediate.OurTypeAnnotation) and isinstance(
            prop_type.our_type, intermediate.ConstrainedPrimitive
        ):
            properties_constrained_primitives.append(prop)
        else:
            properties_non_constrained_primitives.append(prop)

    # endregion

    # region Include constrained primitive verification in non-recusrive part

    # Since constrained primitives are not included in the recursion, we verify
    # properties of constrained primitives at the non-recurse block as well.
    for prop in properties_constrained_primitives:
        prop_type = intermediate.beneath_optional(prop.type_annotation)
        if isinstance(prop_type, intermediate.OurTypeAnnotation) and isinstance(
            prop_type.our_type, intermediate.ConstrainedPrimitive
        ):
            constrained_primitive_block, error = _generate_verify_property_snippet(
                prop=prop
            )
            if error is not None:
                errors.append(error)
            else:
                assert constrained_primitive_block is not None
                if constrained_primitive_block != "":
                    blocks.append(constrained_primitive_block)

    if len(errors) > 0:
        return None, errors

    # endregion

    # region Generate the recursive part

    recurse_prop_blocks = []  # type: List[Stripped]

    for prop in properties_non_constrained_primitives:
        prop_block, error = _generate_verify_property_snippet(prop=prop)
        if error is not None:
            errors.append(error)
        else:
            assert prop_block is not None
            if prop_block != "":
                recurse_prop_blocks.append(prop_block)

    if len(errors) > 0:
        return None, errors

    if len(recurse_prop_blocks) > 0:
        joined_prop_blocks = "\n\n".join(recurse_prop_blocks)
        blocks.append(
            Stripped(
                f"""\
if (context === true) {{
{I}{indent_but_first_line(joined_prop_blocks, I)}
}}"""
            )
        )

    # endregion

    cls_name = typescript_naming.class_name(cls.name)

    no_verification = len(blocks) == 0
    if len(blocks) == 0:
        blocks.append(
            Stripped(
                f"""\
// No verification has been defined for {cls_name}."""
            )
        )

    transform_name = typescript_naming.method_name(
        Identifier(f"transform_{cls.name}_with_context")
    )

    maybe_disable_that_unused = (
        ""
        if not no_verification
        else f"{I}// eslint-disable-next-line @typescript-eslint/no-unused-vars\n"
    )

    maybe_disable_context_unused = (
        ""
        if len(recurse_prop_blocks) > 0
        else f"{I}// eslint-disable-next-line @typescript-eslint/no-unused-vars\n"
    )

    writer = io.StringIO()
    writer.write(
        f"""\
*{transform_name}(
{maybe_disable_that_unused}{I}that: AasTypes.{cls_name},
{maybe_disable_context_unused}{I}context: boolean
): IterableIterator<VerificationError> {{
"""
    )

    for i, stmt in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(stmt, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_transformer(
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a transformer to double-dispatch an instance to errors."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    # The abstract classes are directly dispatched by the transformer,
    # so we do not need to handle them separately.
    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            transform_key = specific_implementations.ImplementationKey(
                f"Verification/transform_{cls.name}.ts"
            )

            implementation = spec_impls.get(transform_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The transformation snippet is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {transform_key}",
                    )
                )
                continue

            blocks.append(spec_impls[transform_key])
        else:
            block, cls_errors = _generate_transform_for_class(
                cls=cls,
                symbol_table=symbol_table,
                base_environment=base_environment,
            )
            if cls_errors is not None:
                errors.extend(cls_errors)
            else:
                assert block is not None
                blocks.append(block)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Verify an instance of the model recursively or non-recursively (depending on the context).
 */
class Verifier
{I}extends AasTypes.AbstractTransformerWithContext<
{II}boolean, IterableIterator<VerificationError>
{I}> {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_verify_constrained_primitive(
    constrained_primitive: intermediate.ConstrainedPrimitive,
    symbol_table: intermediate.SymbolTable,
    base_environment: intermediate_type_inference.Environment,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the verify function for the constrained primitives."""
    errors = []  # type: List[Error]
    blocks = []  # type: List[Stripped]

    environment = intermediate_type_inference.MutableEnvironment(
        parent=base_environment
    )

    assert environment.find(Identifier("self")) is None
    environment.set(
        identifier=Identifier("self"),
        type_annotation=intermediate_type_inference.OurTypeAnnotation(
            our_type=constrained_primitive
        ),
    )

    for invariant in constrained_primitive.invariants:
        invariant_code, error = _transpile_invariant(
            invariant=invariant, symbol_table=symbol_table, environment=environment
        )
        if error is not None:
            errors.append(
                Error(
                    constrained_primitive.parsed.node,
                    f"Failed to transpile the invariant of "
                    f"the constrained primitive {constrained_primitive.name!r}",
                    [error],
                )
            )
            continue

        assert invariant_code is not None

        blocks.append(invariant_code)

    if len(errors) > 0:
        return None, errors

    no_verification = False
    if len(blocks) == 0:
        blocks.append(Stripped("// There is no verification specified."))
        no_verification = True

    function_name = typescript_naming.function_name(
        Identifier(f"verify_{constrained_primitive.name}")
    )

    that_type = typescript_common.PRIMITIVE_TYPE_MAP[constrained_primitive.constrainee]

    writer = io.StringIO()

    writer.write(
        """\
/**
 * Verify the constraints of `that` value.
 *
 * @param that - to be verified
 * @returns errors, if any
 */
"""
    )

    if no_verification:
        writer.write(
            f"""\
export function *{function_name}(
{I}// eslint-disable-next-line @typescript-eslint/no-unused-vars
{I}that: {that_type}
): IterableIterator<VerificationError> {{
"""
        )
    else:
        writer.write(
            f"""\
export function *{function_name}(
{I}that: {that_type}
): IterableIterator<VerificationError> {{
"""
        )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    assert len(errors) == 0
    return Stripped(writer.getvalue()), None


def _generate_module_comment(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the documentation comment for the module."""
    package_identifier_key = specific_implementations.ImplementationKey(
        "package_identifier.txt"
    )

    package_identifier = spec_impls.get(package_identifier_key, None)
    if package_identifier is None:
        return None, Error(
            None,
            f"The package identifier snippet is missing "
            f"in the specific implementations: {package_identifier_key}",
        )

    assert package_identifier is not None

    blocks = [
        Stripped("Verify that the instances of the meta-model satisfy the invariants.")
    ]  # type: List[Stripped]

    first_cls = (
        symbol_table.concrete_classes[0]
        if len(symbol_table.concrete_classes) > 0
        else None
    )  # type: Optional[intermediate.ConcreteClass]

    if first_cls is not None:
        cls_name = typescript_naming.class_name(first_cls.name)
        an_instance_variable = typescript_naming.variable_name(
            Identifier("an_instance")
        )

        blocks.append(
            Stripped(
                f"""\
Here is an example how to verify an instance of {{@link types.{cls_name}}}:

```ts
import * as AasTypes from "{package_identifier}/types";
import * as AasVerification from "{package_identifier}/verification";

const {an_instance_variable} = new AasTypes.{cls_name}(
{I}// ... some constructor arguments ...
);

for (const error of AasVerification.verify({an_instance_variable})) {{
{I}console.log(`${{error.message}} at: ${{error.path}}`);
}}
```"""
            )
        )

    # endregion

    text = "\n\n".join(blocks)

    return Stripped(typescript_description.documentation_comment(Stripped(text))), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the TypeScript code for verification based on the symbol table."""
    errors = []  # type: List[Error]

    module_docstring, module_docstring_error = _generate_module_comment(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if module_docstring_error is not None:
        errors.append(module_docstring_error)
        # NOTE (mristin, 2023-03-18):
        # Allow the execution to continue to catch other errors as well
        module_docstring = Stripped("")

    assert module_docstring is not None

    blocks = [
        module_docstring,
        typescript_common.WARNING,
        Stripped(
            """\
import * as AasCommon from "./common";
import * as AasConstants from "./constants";
import * as AasTypes from "./types";"""
        ),
        Stripped(
            """\
// The generated code might contain deliberately double negations. For example,
// when the constraint is formulated as a NAND and we check that the constraint
// is not fulfilled. Therefore, we disable this linting rule.
/* eslint no-extra-boolean-cast: 0 */"""
        ),
        Stripped(
            f"""\
/**
 * Represent a property access on a path to an erroneous value.
 */
export class PropertySegment {{
{I}/**
{I} * Instance containing the property
{I} */
{I}readonly instance: AasTypes.Class;

{I}/**
{I} * Name of the property
{I} */
{I}readonly name: string;

{I}constructor(instance: AasTypes.Class, name: string) {{
{II}this.instance = instance;
{II}this.name = name;
{I}}}

{I}toString(): string {{
{II}return `.${{this.name}}`;
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent an index access on a path to an erroneous value.
 */
export class IndexSegment {{
{I}/**
{I} * Sequence containing the item at {{@link index}}
{I} */
{I}readonly sequence: Array<AasTypes.Class>;

{I}/**
{I} * Index of the item in the {{@link sequence}}
{I} */
{I}readonly index: number;

{I}constructor(sequence: Array<AasTypes.Class>, index: number) {{
{II}this.sequence = sequence;
{II}this.index = index;
{I}}}

{I}toString(): string {{
{II}return `[${{this.index}}]`;
{I}}}
}}"""
        ),
        Stripped("export type Segment = PropertySegment | IndexSegment;"),
        Stripped(
            f"""\
/**
 * Represent the relative path to the erroneous value.
 */
export class Path {{
{I}readonly segments: Array<Segment> = [];

{I}prepend(segment: Segment): void {{
{II}this.segments.unshift(segment);
{I}}}

{I}toString(): string {{
{II}return this.segments.join("");
{I}}}
}}"""
        ),
        Stripped(
            f"""\
/**
 * Represent a verification error in the data.
 */
export class VerificationError {{
{I}// NOTE (mristin, 2022-11-12):
{I}// The name `VerificationError` is redundant since it lives in `verification` module,
{I}// and it would have made more sense to call it simply `Error`. Unfortunately in this case,
{I}// `Error` is a reserved name by JavaScript.

{I}/**
{I} * Human-readable description of the error
{I} */
{I}readonly message: string;

{I}/**
{I} * Path to the erroneous value
{I} */
{I}readonly path: Path = new Path();

{I}/**
{I} * Initialize with the given `message` and `path`.
{I} *
{I} * @remarks
{I} * If no `path` is specified, initialize with an empty path.
{I} */
{I}constructor(message: string, path: Path | null = null) {{
{II}this.message = message;
{II}this.path = (path !== null)
{III}? path
{III}: new Path();
{I}}}
}}"""
        ),
    ]  # type: List[Stripped]

    base_environment = intermediate_type_inference.populate_base_environment(
        symbol_table=symbol_table
    )

    for verification in symbol_table.verification_functions:
        if isinstance(verification, intermediate.ImplementationSpecificVerification):
            implementation_key = specific_implementations.ImplementationKey(
                f"Verification/{verification.name}.ts"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        None,
                        f"The snippet for the verification function "
                        f"{verification.name!r} is missing: {implementation_key}",
                    )
                )
            else:
                blocks.append(implementation)

        elif isinstance(verification, intermediate.PatternVerification):
            implementation, error = _transpile_pattern_verification(
                verification=verification
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        elif isinstance(verification, intermediate.TranspilableVerification):
            implementation, error = _transpile_transpilable_verification(
                verification=verification,
                symbol_table=symbol_table,
                environment=base_environment,
            )

            if error is not None:
                errors.append(error)
            else:
                assert implementation is not None
                blocks.append(implementation)

        else:
            assert_never(verification)

    transformer_block, transformer_errors = _generate_transformer(
        symbol_table=symbol_table,
        base_environment=base_environment,
        spec_impls=spec_impls,
    )
    if transformer_errors is not None:
        errors.extend(transformer_errors)
    else:
        assert transformer_block is not None
        blocks.append(transformer_block)

    blocks.append(Stripped("const VERIFIER = new Verifier();"))

    blocks.append(
        Stripped(
            f"""\
/**
 * Verify the constraints of `that`.
 *
 * @param that - instance to be verified
 * @param recurse - if set, continue the verification recursively
 * @returns a stream of verification errors
 */
export function *verify(
  that: AasTypes.Class,
  recurse = true
): IterableIterator<VerificationError> {{
{I}yield * VERIFIER.transformWithContext(that, recurse);
}}"""
        )
    )

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            # NOTE (mristin, 2022-11-12):
            # We do not verify the enumerations explicitly in TypeScript and
            # leave those checks to TypeScript compiler.
            pass

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            (
                constrained_primitive_block,
                constrained_primitive_errors,
            ) = _generate_verify_constrained_primitive(
                constrained_primitive=our_type,
                symbol_table=symbol_table,
                base_environment=base_environment,
            )

            if constrained_primitive_errors is not None:
                errors.extend(constrained_primitive_errors)
            else:
                assert constrained_primitive_block is not None
                blocks.append(constrained_primitive_block)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2022-11-12):
            # We provide a general dispatch function for the most abstract
            # class ``Class``.
            pass
        else:
            assert_never(our_type)

    blocks.append(typescript_common.WARNING)

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate the test code for the XML de/serialization of interfaces."""

from typing import List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, indent_but_first_line
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


# fmt: off
@ensure(
    lambda result: result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    namespace: csharp_common.NamespaceIdentifier,
    symbol_table: intermediate.SymbolTable,
) -> str:
    """
    Generate the test code for the XML de/serialization of interfaces.

    The ``namespace`` indicates the fully-qualified name of the base project.
    """
    blocks = []  # type: List[Stripped]

    for cls in symbol_table.classes:
        if cls.interface is None or len(cls.interface.implementers) == 0:
            continue

        for implementer_cls in cls.interface.implementers:
            if (
                implementer_cls.serialization is None
                or not implementer_cls.serialization.with_model_type
            ):
                continue

            interface_name_csharp = csharp_naming.interface_name(cls.interface.name)

            implementer_cls_name_csharp = csharp_naming.class_name(implementer_cls.name)

            blocks.append(
                Stripped(
                    f"""\
[Test]
public void Test_round_trip_{interface_name_csharp}_from_{implementer_cls_name_csharp}()
{{
{I}// We load from JSON here just to jump-start the round trip.
{I}// The round-trip goes then over XML.
{I}var instance = Aas.Tests.CommonJsonization.LoadMaximal{implementer_cls_name_csharp}();

{I}// The round-trip starts here.
{I}var outputBuilder = new System.Text.StringBuilder();

{I}// Serialize to XML
{I}{{
{II}using var xmlWriter = System.Xml.XmlWriter.Create(
{III}outputBuilder,
{III}new System.Xml.XmlWriterSettings()
{III}{{
{IIII}Encoding = System.Text.Encoding.UTF8,
{IIII}OmitXmlDeclaration = true
{III}}});

{II}Aas.Xmlization.Serialize.To(
{III}instance,
{III}xmlWriter);
{I}}}

{I}// De-serialize from XML
{I}string outputText = outputBuilder.ToString();

{I}using var outputReader = new System.IO.StringReader(outputText);

{I}using var xmlReader = System.Xml.XmlReader.Create(
{II}outputReader,
{II}new System.Xml.XmlReaderSettings());

{I}var anotherInstance = Aas.Xmlization.Deserialize.{interface_name_csharp}From(
{II}xmlReader);

{I}// Serialize back to XML
{I}var anotherOutputBuilder = new System.Text.StringBuilder();

{I}{{
{II}using var anotherXmlWriter = System.Xml.XmlWriter.Create(
{III}anotherOutputBuilder,
{III}new System.Xml.XmlWriterSettings()
{III}{{
{IIII}Encoding = System.Text.Encoding.UTF8,
{IIII}OmitXmlDeclaration = true
{III}}});

{II}Aas.Xmlization.Serialize.To(
{III}anotherInstance,
{III}anotherXmlWriter);
{I}}}

{I}// Compare
{I}Assert.AreEqual(outputText, anotherOutputBuilder.ToString());
}}  // void Test_round_trip_{interface_name_csharp}_from_{implementer_cls_name_csharp}"""
                )
            )

    blocks_joined = "\n\n".join(blocks)

    return f"""\
{csharp_common.WARNING}

using Aas = {namespace};  // renamed

using NUnit.Framework;  // can't alias

namespace {namespace}.Tests
{{
{I}public class TestXmlizationOfInterfaces
{I}{{
{II}{indent_but_first_line(blocks_joined, II)}
{I}}}  // class TestXmlizationOfInterfaces
}}  // namespace {namespace}.Tests

{csharp_common.WARNING}
"""


assert generate.__doc__ is not None
assert generate.__doc__.strip().startswith(__doc__.strip())



"""Generate Python code of common functions by including the code directly."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.python import (
    common as python_common,
)
from aas_core_codegen.python.common import (
    INDENT as I,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the Python code for common functions."""
    blocks = [
        Stripped('"""Provide common functions shared among the modules."""'),
        python_common.WARNING,
        Stripped(
            f"""\
from typing import (
{I}NoReturn
)"""
        ),
        Stripped(
            f"""\
def assert_never(value: NoReturn) -> NoReturn:
{I}\"\"\"
{I}Signal to mypy to perform an exhaustive matching.

{I}Please see the following page for more details:
{I}https://hakibenita.com/python-mypy-exhaustive-checking
{I}\"\"\"
{I}assert False, f"Unhandled value: {{value}} ({{type(value).__name__}})\""""
        ),
        python_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Generate the C# constants corresponding to the constants of the meta-model."""
import io
import textwrap
from typing import (
    Optional,
    List,
    Tuple,
)

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.csharp import (
    common as csharp_common,
    naming as csharp_naming,
)
from aas_core_codegen.csharp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)


def _byte_array_as_expr(value: bytearray) -> Stripped:
    """Return a multi-line C# code representing the ``value``."""
    writer = io.StringIO()

    writer.write("new byte[]\n{\n")
    for i, byte in enumerate(value):
        if i > 0:
            if i % 8 == 0:
                writer.write(",\n")
            else:
                writer.write(", ")

        writer.write(f"0x{byte:02x}")

    writer.write("\n}")

    return Stripped(writer.getvalue())


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_primitive(
    constant: intermediate.ConstantPrimitive,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant primitive."""
    constant_name = csharp_naming.property_name(constant.name)

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        literal = "true" if constant.value else "false"

        return Stripped(f"public static const bool {constant_name} = {literal};"), None

    elif constant.a_type is intermediate.PrimitiveType.INT:
        literal = str(constant.value)

        return Stripped(f"public static const long {constant_name} = {literal};"), None

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (mristin, 2022-07-06):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.
        literal = str(constant.value)

        return (
            Stripped(f"public static const double {constant_name} = {literal};"),
            None,
        )

    elif constant.a_type is intermediate.PrimitiveType.STR:
        assert isinstance(constant.value, str)
        literal = csharp_common.string_literal(constant.value)

        return (
            Stripped(f"public static const string {constant_name} = {literal};"),
            None,
        )

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        assert isinstance(constant.value, bytearray)

        literal = _byte_array_as_expr(value=constant.value)

        return (
            Stripped(
                f"""\
public static readonly byte[] {constant_name} = (
{I}{indent_but_first_line(literal, I)});"""
            ),
            None,
        )

    else:
        assert_never(constant.a_type)


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_primitives(
    constant: intermediate.ConstantSetOfPrimitives,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of primitives."""
    constant_name = csharp_naming.property_name(constant.name)

    writer = io.StringIO()

    if constant.a_type is intermediate.PrimitiveType.BOOL:
        writer.write(
            f"""\
public static readonly HashSet<bool> {constant_name} = (
{I}new HashSet<bool>()
{I}{{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent("true" if literal.value else "false", II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"{I}}});")

    elif constant.a_type is intermediate.PrimitiveType.INT:
        writer.write(
            f"""\
public static readonly HashSet<long> {constant_name} = (
{I}new HashSet<long>()
{I}{{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"{I}}});")

    elif constant.a_type is intermediate.PrimitiveType.FLOAT:
        # NOTE (mristin, 2022-07-06):
        # We assume that the float constants are not meant to be all to precise.
        # Therefore, we use a string representation here. However, beware that we
        # might have to use a more precise representation in the future if the spec
        # change.

        writer.write(
            f"""\
public static readonly HashSet<double> {constant_name} = (
{I}new HashSet<double>()
{I}{{
"""
        )

        for i, literal in enumerate(constant.literals):
            writer.write(textwrap.indent(str(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"{I}}});")

    elif constant.a_type is intermediate.PrimitiveType.STR:
        writer.write(
            f"""\
public static readonly HashSet<string> {constant_name} = (
{I}new HashSet<string>()
{I}{{
"""
        )

        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, str)

            writer.write(
                textwrap.indent(csharp_common.string_literal(literal.value), II)
            )

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"{I}}});")

    elif constant.a_type is intermediate.PrimitiveType.BYTEARRAY:
        writer.write(
            f"""\
public static readonly HashSet<byte[]> {constant_name} = (
{I}new HashSet<byte[]>(new ByteArrayComparer())
{I}{{
"""
        )

        for i, literal in enumerate(constant.literals):
            assert isinstance(literal.value, bytearray)

            writer.write(textwrap.indent(_byte_array_as_expr(literal.value), II))

            if i < len(constant.literals) - 1:
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"{I}}});")

    else:
        assert_never(constant.a_type)

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_constant_set_of_enumeration_literals(
    constant: intermediate.ConstantSetOfEnumerationLiterals,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of a constant set of enumeration literals."""
    constant_name = csharp_naming.property_name(constant.name)
    enum_name = csharp_naming.enum_name(constant.enumeration.name)

    writer = io.StringIO()

    # NOTE (mristin, 2022-08-19):
    # We make the sets of enumeration literals work on nullables to avoid checking
    # nullability all the time in the code. This gives a bit less performant code,
    # but a much more readable one.

    writer.write(
        f"""\
public static readonly HashSet<{enum_name}?> {constant_name} = (
{I}new HashSet<{enum_name}?>()
{I}{{
"""
    )

    for i, literal in enumerate(constant.literals):
        literal_name = csharp_naming.enum_literal_name(literal.name)

        writer.write(textwrap.indent(f"{enum_name}.{literal_name}", II))

        if i < len(constant.literals) - 1:
            writer.write(",\n")
        else:
            writer.write("\n")

    writer.write(f"{I}}});")

    return Stripped(writer.getvalue()), None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    namespace: csharp_common.NamespaceIdentifier,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code of the constants based on the symbol table.

    The ``namespace`` defines the AAS C# namespace.
    """
    constants_blocks = []  # type: List[Stripped]

    needs_byte_array_comparer = False
    for constant in symbol_table.constants:
        if (
            isinstance(constant, intermediate.ConstantSetOfPrimitives)
            and constant.a_type is intermediate.PrimitiveType.BYTEARRAY
        ):
            needs_byte_array_comparer = True
            break

    if needs_byte_array_comparer:
        constants_blocks.append(
            Stripped(
                f"""\
/// <summary>
/// Compare two byte arrays for equality.
/// </summary>
/// <remarks>
/// Based on: https://stackoverflow.com/a/49739848/1600678
/// </remarks>
private class ByteArrayComparer
{{
{I}public bool Equals(byte[] a, byte[] b)
{I}{{
{II}if (a.Length != b.Length) return false;
{II}for (int i = 0; i < a.Length; i++)
{II}{{
{III}if (a[i] != b[i]) return false;
{II}}}

{II}return true;
{I}}}

{I}public int GetHashCode(byte[] a)
{I}{{
{II}uint b = 0;
{II}for (int i = 0; i < a.Length; i++)
{II}{{
{III}b = ((b << 23) | (b >> 9)) ^ a[i];
{II}}}

{II}return unchecked((int)b);
{I}}}
}}"""
            )
        )

    errors = []  # type: List[Error]

    for constant in symbol_table.constants:
        constants_block: Optional[Stripped]
        error: Optional[Error]

        if isinstance(constant, intermediate.ConstantPrimitive):
            constants_block, error = _generate_constant_primitive(constant=constant)
        elif isinstance(constant, intermediate.ConstantSetOfPrimitives):
            constants_block, error = _generate_constant_set_of_primitives(
                constant=constant
            )
        elif isinstance(constant, intermediate.ConstantSetOfEnumerationLiterals):
            constants_block, error = _generate_constant_set_of_enumeration_literals(
                constant=constant
            )
        else:
            assert_never(constant)

        if error is not None:
            errors.append(error)
            continue

        assert constants_block is not None
        constants_blocks.append(constants_block)

    if len(errors) > 0:
        return None, errors

    constants_writer = io.StringIO()
    constants_writer.write(
        """\
/// <summary>
/// Provide constant values of the meta-model.
/// </summary>
public static class Constants
{
"""
    )

    for i, constants_block in enumerate(constants_blocks):
        assert len(constants_block) > 0

        if i > 0:
            constants_writer.write("\n\n")

        constants_writer.write(textwrap.indent(constants_block, I))

    constants_writer.write("\n}  // public static class Constants")

    blocks = [
        csharp_common.WARNING,
        Stripped("using System.Collections.Generic;  // can't alias"),
        Stripped(
            f"""\
namespace {namespace}
{{
{I}{indent_but_first_line(constants_writer.getvalue(), I)}
}}  // namespace {namespace}"""
        ),
        csharp_common.WARNING,
    ]  # type: List[Stripped]

    out = io.StringIO()
    for i, constants_block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(constants_block)

    out.write("\n")

    return out.getvalue(), None



"""Generate builders for the AAS data structures."""

import io
import textwrap
from typing import (
    List,
    Optional,
    Tuple,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen.common import (
    Error,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.java import (
    common as java_common,
    description as java_description,
    naming as java_naming,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
)


# fmt: off
@require(
    lambda cls: any(
        isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation)
        for arg in cls.constructor.arguments
    )
)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def _generate_builder(
    cls: intermediate.ConcreteClass,
    package: java_common.PackageIdentifier,
) -> Tuple[Optional[java_common.JavaFile], Optional[Error]]:

    class_name = java_naming.class_name(cls.name)

    builder_name = f"{class_name}Builder"

    file_name = f"{builder_name}.java"

    package_name = java_common.PackageIdentifier(f"{package}.generation")

    builder_blocks = []  # type: List[Stripped]

    # region Properties

    properties_blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        prop_type = java_common.generate_type(type_annotation=type_anno)

        prop_name = java_naming.property_name(prop.name)

        blocks = []  # type: List[Stripped]

        if prop.description is not None:
            (
                prop_comment,
                prop_comment_errors,
            ) = java_description.generate_comment_for_property(
                description=prop.description,
                context=java_description.Context(
                    package=package,
                    # NOTE (mristin, 2024-03-27):
                    # Though the builder is associated with a class, the links in
                    # the description need to reference the original class, so
                    # we do not indicate the class for the context here.
                    cls_or_enum=None,
                ),
            )

            if prop_comment_errors:
                return None, Error(
                    prop.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for the property {prop.name!r}",
                    prop_comment_errors,
                )

            assert prop_comment is not None

            blocks.append(prop_comment)

        blocks.append(
            Stripped(
                f"""\
private {prop_type} {prop_name};"""
            )
        )

        properties_blocks.append(Stripped("\n".join(blocks)))

    builder_blocks.append(Stripped("\n\n".join(properties_blocks)))

    # endregion

    # region Constructor

    arg_codes = []  # type: List[Stripped]

    for arg in cls.constructor.arguments:
        if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        arg_type = java_common.generate_type(type_annotation=arg.type_annotation)

        arg_name = java_naming.argument_name(arg.name)

        arg_codes.append(Stripped(f"{arg_type} {arg_name}"))

    if len(arg_codes) > 0:

        constructor_writer = io.StringIO()

        if len(arg_codes) == 1:
            constructor_writer.write(f"public {builder_name}({arg_codes[0]}) {{\n")
        else:
            arg_block = ",\n".join(arg_codes)
            arg_block_indented = textwrap.indent(arg_block, I)
            constructor_writer.write(
                f"public {builder_name}(\n{arg_block_indented}) {{\n"
            )

        for arg in cls.constructor.arguments:
            if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
                continue

            arg_name = java_naming.argument_name(arg.name)

            constructor_writer.write(
                f"""\
{I}this.{arg_name} = Objects.requireNonNull(
{II}{arg_name},
{II}"Argument \\"{arg_name}\\" must be non-null.");
"""
            )

        constructor_writer.write("}")

        constructor = constructor_writer.getvalue()

        builder_blocks.append(Stripped(constructor))

    # endregion

    # region Setters

    setter_blocks = []  # type: List[Stripped]

    for arg in cls.constructor.arguments:
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        type_anno = intermediate.beneath_optional(arg.type_annotation)

        arg_type = java_common.generate_type(type_annotation=type_anno)

        arg_name = java_naming.argument_name(arg.name)

        setter_name = java_naming.setter_name(arg.name)

        setter_blocks.append(
            Stripped(
                f"""\
public {builder_name} {setter_name}({arg_type} {arg_name}) {{
{I}this.{arg_name} = {arg_name};
{I}return this;
}}"""
            )
        )

    builder_blocks.append(Stripped("\n\n".join(setter_blocks)))

    # endregion

    # region Build method

    build_method_writer = io.StringIO()

    build_method_writer.write(f"public {class_name} build() {{\n")

    build_method_writer.write(f"{I}return new {class_name}(")

    if len(cls.constructor.arguments) == 1:
        arg_name = java_naming.argument_name(cls.constructor.arguments[0].name)

        build_method_writer.write(f"this.{arg_name}")
    else:
        for idx, arg in enumerate(cls.constructor.arguments):
            arg_name = java_naming.argument_name(arg.name)

            if idx >= 1:
                build_method_writer.write(",")

            build_method_writer.write(f"\n{II}this.{arg_name}")

    build_method_writer.write(");")

    build_method_writer.write("\n}")

    builder_blocks.append(Stripped(build_method_writer.getvalue()))

    # endregion

    builder = Stripped("\n\n".join(builder_blocks))

    return (
        java_common.JavaFile(
            name=file_name,
            content=f"""\
package {package_name};

import {package}.types.enums.*;
import {package}.types.impl.*;
import {package}.types.model.*;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

/**
 * Builder for the {class_name} type.
 */
public class {builder_name} {{
{I}{indent_but_first_line(builder, I)}
}}
""",
        ),
        None,
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
) -> Tuple[Optional[List[java_common.JavaFile]], Optional[List[Error]]]:
    """
    Generate the Java code of the structures based on the symbol table.

    The ``package`` defines the root Java package.
    """

    files = []  # type: List[java_common.JavaFile]
    errors = []  # type: List[Error]

    for our_type in symbol_table.our_types:
        if not isinstance(
            our_type,
            intermediate.ConcreteClass,
        ):
            continue

        if not any(
            isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation)
            for arg in our_type.constructor.arguments
        ):
            continue

        new_file, error = _generate_builder(our_type, package)

        if error is not None:
            errors.append(error)
            continue

        assert new_file is not None

        files.append(new_file)

    if len(errors) > 0:
        return None, errors

    return files, None



"""Generate the code to test descend and pass-through visitor jointly."""

import io

from icontract import ensure

from aas_core_codegen.common import Stripped
from aas_core_codegen.python import common as python_common
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(qualified_module_name: python_common.QualifiedModuleName) -> str:
    """
    Generate the code to test descend and pass-through visitor jointly.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            f'''\
"""
Test jointly :py:method:`{qualified_module_name}.types.Class.descend` and
:py:method:`{qualified_module_name}.types.PassThroughVisitor`.
"""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            f"""\
from typing import (
{I}List,
{I}Sequence
)
import unittest"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.types as aas_types"""
        ),
        Stripped(
            """\
import tests.common_xmlization"""
        ),
        Stripped(
            f'''\
class _TracingVisitor(aas_types.PassThroughVisitor):
{I}"""Visit the instances and trace them."""

{I}def __init__(self) -> None:
{II}"""Initialize with an empty log."""
{II}self._log = []  # type: List[str]

{I}@property
{I}def log(self) -> Sequence[str]:
{II}"""Get the tracing log."""
{II}return self._log

{I}def visit(self, that: aas_types.Class) -> None:
{II}self._log.append(tests.common.trace(that))
{II}super().visit(that)'''
        ),
        Stripped(
            f'''\
def assert_tracing_logs_from_descend_and_visitor_are_the_same(
{I}that: aas_types.Class, test_case: unittest.TestCase
) -> None:
{I}"""
{I}Check that the tracing logs are the same when :paramref:`that` instance
{I}is visited and when :paramref:`that` is ran through
{I}:py:method:`aas_types.Class.descend`.

{I}:param that: instance to be iterated over
{I}:param test_case: in which this assertion runs
{I}:raise: :py:class:`AssertionError` if the logs differ
{I}"""
{I}log_from_descend = [tests.common.trace(something) for something in that.descend()]

{I}visitor = _TracingVisitor()
{I}visitor.visit(that)
{I}log_from_visitor = visitor.log

{I}test_case.assertGreater(len(log_from_visitor), 0)
{I}test_case.assertEqual(tests.common.trace(that), log_from_visitor[0])

{I}# noinspection PyTypeChecker
{I}test_case.assertListEqual(log_from_descend, log_from_visitor[1:])  # type: ignore'''
        ),
        Stripped(
            f"""\
class TestWithExpectedMaximal(unittest.TestCase):
{I}def test_descend_against_recorded_trace_log(self) -> None:
{II}xml_expected_dir = tests.common.TEST_DATA_DIR / "Xml" / "Expected"

{II}for path in sorted(xml_expected_dir.glob("**/maximal.xml")):
{III}instance = tests.common_xmlization.must_load(path)

{III}rel_path = path.relative_to(xml_expected_dir)

{III}expected_path = (
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "descend_and_pass_through_visitor"
{IIII}/ rel_path.parent
{IIII}/ f"{{rel_path.name}}.trace"
{III})

{III}log = [
{IIII}tests.common.trace(something)
{IIII}for something in instance.descend()
{III}]

{III}got_text = tests.common.trace_log_as_text_file_content(log)

{III}tests.common.record_or_check(expected_path, got_text)

{I}def test_descend_against_visitor(self) -> None:
{II}xml_expected_dir = tests.common.TEST_DATA_DIR / "Xml" / "Expected"

{II}for path in sorted(xml_expected_dir.glob("**/maximal.xml")):
{III}instance = tests.common_xmlization.must_load(path)

{III}assert_tracing_logs_from_descend_and_visitor_are_the_same(
{IIII}instance,
{IIII}self
{III})"""
        ),
        Stripped(
            f"""\
if __name__ == "__main__":
{I}unittest.main()"""
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Generate the test code for the ``X_or_default`` methods."""

import io
from typing import List, Optional

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_test_case(cls: intermediate.ConcreteClass) -> Optional[Stripped]:
    """
    Generate the test case class for the given ``cls``.

    If there are no ``X_or_default`` methods, return None.
    """
    xml_cls_name = naming.xml_class_name(cls.name)

    python_cls_name = python_naming.class_name(cls.name)

    test_methods = []  # type: List[Stripped]

    for method in cls.methods:
        if not method.name.endswith("_or_default"):
            continue

        if method.returns is None:
            continue

        test_case_method_name = python_naming.method_name(
            Identifier(f"test_{method.name}_against_recorded")
        )

        x_or_default_name = python_naming.method_name(method.name)

        file_name = f"{x_or_default_name}.trace"

        test_methods.append(
            Stripped(
                f"""\
def {test_case_method_name}(self) -> None:
{I}for minimal_or_maximal in ["minimal", "maximal"]:
{II}instance = tests.common_xmlization.must_load(
{III}pathlib.Path(
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "Xml"
{IIII}/ "Expected"
{IIII}/ {xml_cls_name!r}
{IIII}/ f"{{minimal_or_maximal}}.xml"
{III})
{II})

{II}assert isinstance(
{III}instance,
{III}aas_types.{python_cls_name}
{II})

{II}log = [tests.common.trace(instance.{x_or_default_name}())]

{II}got_text = tests.common.trace_log_as_text_file_content(log)

{II}expected_path = pathlib.Path(
{III}tests.common.TEST_DATA_DIR
{III}/ "test_X_or_default"
{III}/ {xml_cls_name!r}
{III}/ f"on_{{minimal_or_maximal}}.xml"
{III}/ {file_name!r}
{II})

{II}tests.common.record_or_check(expected_path, got_text)"""
            )
        )

    if len(test_methods) == 0:
        return None

    test_class_name = python_naming.class_name(Identifier(f"Test_{cls.name}"))

    body = "\n\n".join(test_methods)

    return Stripped(
        f"""\
class {test_class_name}(unittest.TestCase):
{I}{indent_but_first_line(body, I)}"""
    )


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> str:
    """
    Generate the test code for the ``X_or_default`` methods.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '''\
"""Test ``X_or_default`` methods."""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import pathlib
import unittest"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.types as aas_types"""
        ),
        Stripped(
            """\
import tests.common_xmlization"""
        ),
    ]

    for cls in symbol_table.concrete_classes:
        test_case = _generate_test_case(cls=cls)

        if test_case is None:
            continue

        blocks.append(test_case)

    blocks.extend(
        [
            Stripped(
                f"""\
if __name__ == "__main__":
{I}unittest.main()"""
            ),
            python_common.WARNING,
        ]
    )

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Generate the Golang data structures from the intermediate representation."""
import io
import itertools
import textwrap
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    cast,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen import specific_implementations
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.golang import (
    common as golang_common,
    naming as golang_naming,
    description as golang_description,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
)


# region Checks


def _human_readable_identifier(
    something: Union[
        intermediate.Enumeration,
        intermediate.AbstractClass,
        intermediate.ConcreteClass,
        intermediate.EnumerationLiteral,
    ]
) -> str:
    """
    Represent ``something`` in a human-readable text.

    The reader should be able to trace ``something`` back to the meta-model.
    """
    # NOTE (mristin, 2023-03-29):
    # This function has been copy-pasted from
    # :py:mod:`aas_core_codegen.python.structure._generate`. We tried to refactor it to
    # :py:mod:`aas_core_codegen.intermediate`, but it turned out that the refactored
    # code was nigh unreadable. So we preferred a little bit of copying to a little
    # bit of complexity.

    result: str

    if isinstance(something, intermediate.Enumeration):
        result = f"meta-model enumeration {something.name!r}"
    elif isinstance(something, intermediate.EnumerationLiteral):
        result = f"meta-model enumeration literal {something.name!r}"
    elif isinstance(something, intermediate.AbstractClass):
        result = f"meta-model abstract class {something.name!r}"
    elif isinstance(something, intermediate.ConcreteClass):
        result = f"meta-model concrete class {something.name!r}"
    else:
        assert_never(something)

    assert result is not None
    return result


def _verify_intra_structure_collisions(
    our_type: intermediate.OurType,
) -> Optional[Error]:
    """Verify that no member names collide in the Golang structure of our type."""
    errors = []  # type: List[Error]

    if isinstance(our_type, intermediate.Enumeration):
        # NOTE (mristin, 2023-03-29):
        # We already checked for collisions of enumeration literals in
        # the inter-structure collision checks.
        pass
    elif isinstance(our_type, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(our_type, intermediate.Class):
        observed_member_names = {}  # type: Dict[Identifier, str]

        for prop in our_type.properties:
            name = golang_naming.getter_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"Golang getter {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"Golang getter {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

            name = golang_naming.setter_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"Golang setter {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"Golang setter {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

        for method in our_type.methods:
            method_name = golang_naming.method_name(method.name)

            if method_name in observed_member_names:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"Golang method {method_name!r} corresponding "
                        f"to the meta-model method {method.name!r} collides with "
                        f"the {observed_member_names[method_name]}",
                    )
                )
            else:
                observed_member_names[method_name] = (
                    f"Golang method {method_name!r} corresponding to "
                    f"the meta-model method {method.name!r}"
                )

    else:
        assert_never(our_type)

    if len(errors) > 0:
        errors.append(
            Error(
                our_type.parsed.node,
                f"Naming collision(s) in Golang code for our type {our_type.name!r}",
                underlying=errors,
            )
        )

    return None


def _verify_structure_name_collisions(
    symbol_table: intermediate.SymbolTable,
) -> List[Error]:
    """Verify that the Golang names of the structures do not collide."""
    observed_structure_names: Dict[
        Identifier,
        Union[
            intermediate.Enumeration,
            intermediate.AbstractClass,
            intermediate.ConcreteClass,
            intermediate.EnumerationLiteral,
        ],
    ] = dict()

    errors = []  # type: List[Error]

    # region Inter-structure collisions

    for enum_or_cls in itertools.chain(symbol_table.enumerations, symbol_table.classes):
        names: List[Identifier]

        if isinstance(enum_or_cls, intermediate.Enumeration):
            names = [golang_naming.enum_name(enum_or_cls.name)]
        elif isinstance(enum_or_cls, intermediate.AbstractClass):
            names = [golang_naming.interface_name(enum_or_cls.name)]
        elif isinstance(enum_or_cls, intermediate.ConcreteClass):
            names = [
                golang_naming.interface_name(enum_or_cls.name),
                golang_naming.struct_name(enum_or_cls.name),
            ]
        else:
            assert_never(enum_or_cls)

        for name in names:
            other = observed_structure_names.get(name, None)

            if other is not None:
                errors.append(
                    Error(
                        enum_or_cls.parsed.node,
                        f"The Golang name {name!r} "
                        f"of the {_human_readable_identifier(enum_or_cls)} "
                        f"collides with the Golang name "
                        f"of the {_human_readable_identifier(other)}",
                    )
                )
            else:
                observed_structure_names[name] = enum_or_cls

    # NOTE (mristin, 2023-03-29):
    # Enumeration literals are global constants in Go, so we have to consider them
    # for collisions as well.
    for enum in symbol_table.enumerations:
        for literal in enum.literals:
            name = golang_naming.enum_literal_name(enum.name, literal.name)

            other = observed_structure_names.get(name, None)

            if other is not None:
                errors.append(
                    Error(
                        literal.parsed.node,
                        f"The Golang name {name!r} "
                        f"of the {_human_readable_identifier(literal)} "
                        f"collides with the Golang name "
                        f"of the {_human_readable_identifier(other)}",
                    )
                )
            else:
                observed_structure_names[name] = literal

    # endregion

    # region Intra-structure collisions

    for our_type in symbol_table.our_types:
        collision_error = _verify_intra_structure_collisions(our_type=our_type)

        if collision_error is not None:
            errors.append(collision_error)

    # endregion

    return errors


class VerifiedIntermediateSymbolTable(intermediate.SymbolTable):
    """Represent a verified symbol table which can be used for code generation."""

    # noinspection PyInitNewSignature
    def __new__(
        cls, symbol_table: intermediate.SymbolTable
    ) -> "VerifiedIntermediateSymbolTable":
        raise AssertionError("Only for type annotation")


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def verify(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[VerifiedIntermediateSymbolTable], Optional[List[Error]]]:
    """Verify that Golang code can be generated from the ``symbol_table``."""
    errors = []  # type: List[Error]

    structure_name_collisions = _verify_structure_name_collisions(
        symbol_table=symbol_table
    )

    errors.extend(structure_name_collisions)

    if len(errors) > 0:
        return None, errors

    return cast(VerifiedIntermediateSymbolTable, symbol_table), None


# endregion

# region Generation


@require(lambda enumeration, literal: id(literal) in enumeration.literal_id_set)
@require(lambda literal: literal.description is not None)
def _generate_comment_for_enumeration_literal(
    enumeration: intermediate.Enumeration,
    literal: intermediate.EnumerationLiteral,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    # NOTE (mristin, 2023-03-29):
    # We need to state the pre-condition for the second time for mypy.
    assert literal.description is not None

    # fmt: off
    comment, errors = (
        golang_description.generate_comment_for_summary_remarks(
            description=literal.description,
            context=golang_description.Context(
                package=golang_common.TYPES_PACKAGE,
                cls_or_enum=enumeration
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@require(lambda cls_or_enum: cls_or_enum.description is not None)
def _generate_comment_for_cls_or_enum(
    cls_or_enum: Union[intermediate.Enumeration, intermediate.ClassUnion],
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for our type."""
    # NOTE (mristin, 2023-03-29):
    # We need to state the pre-condition for the second time for mypy.
    assert cls_or_enum.description is not None

    # fmt: off
    comment, errors = (
        golang_description
        .generate_comment_for_summary_remarks_constraints(
            description=cls_or_enum.description,
            context=golang_description.Context(
                package=golang_common.TYPES_PACKAGE, cls_or_enum=None
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_enum(
    enum: intermediate.Enumeration,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the Golang code for the enum."""
    writer = io.StringIO()

    errors = []  # type: List[Error]

    comment = None  # type: Optional[Stripped]
    if enum.description is not None:
        comment, comment_errors = _generate_comment_for_cls_or_enum(cls_or_enum=enum)

        if comment_errors:
            errors.append(
                Error(
                    enum.description.parsed.node,
                    f"Failed to generate the comment "
                    f"for the enumeration {enum.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None

    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    name = golang_naming.enum_name(enum.name)

    writer.write(f"type {name} int;")
    if len(enum.literals) > 0:
        writer.write("const (\n")

        for i, literal in enumerate(enum.literals):
            if literal.description is not None:
                comment, comment_errors = _generate_comment_for_enumeration_literal(
                    enumeration=enum, literal=literal
                )
                if comment_errors is not None:
                    errors.append(
                        Error(
                            literal.description.parsed.node,
                            f"Failed to generate the documentation comment "
                            f"for enumeration literal {literal.name!r}",
                            comment_errors,
                        )
                    )
                else:
                    assert comment is not None
                    writer.write(textwrap.indent(comment, I))
                    writer.write("\n")

            literal_name = golang_naming.enum_literal_name(enum.name, literal.name)

            # NOTE (mristin, 2023-03-29):
            # We optimize for comparisons instead of stringification.
            # The stringification is delegated to a separate module.

            if i == 0:
                writer.write(textwrap.indent(f"{literal_name} {name} = iota", I))
            else:
                writer.write(textwrap.indent(literal_name, I))

            writer.write("\n")

        writer.write(")")

    if len(errors) > 0:
        return None, Error(
            enum.parsed.node,
            f"Failed to generate the Golang code for the enumeration {enum.name!r}",
            errors,
        )

    return Stripped(writer.getvalue()), None


def _literals_of_enum(enum: intermediate.Enumeration) -> Stripped:
    """Generate the Golang code for the constant array listing the literals."""
    name = golang_naming.enum_name(enum.name)
    array_name = golang_naming.constant_name(Identifier(f"literals_of_{enum.name}"))

    literals = []
    for literal in enum.literals:
        literals.append(golang_naming.enum_literal_name(enum.name, literal.name))

    literals_joined = "\n".join(f"{literal}," for literal in literals)

    return Stripped(
        f"""\
// List the literals of [{name}].
//
// Golang does not provide an elegant way to iterate over the literals, so
// this array helps you avoid common errors and pitfalls.
//
// Please do not modify the array in the caller's code.
var {array_name} = [...]{name} {{
{I}{indent_but_first_line(literals_joined, I)}
}}"""
    )


def _generate_definition_for_model_type(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """
    Generate the enumeration corresponding to the model types.

    This is necessary so that we can have fast switches on instance types.
    """
    enum_name = golang_naming.enum_name(Identifier("Model_type"))

    literal_specs = []  # type: List[Stripped]
    for i, cls in enumerate(symbol_table.concrete_classes):
        literal_name = golang_naming.enum_literal_name(
            Identifier("Model_type"), cls.name
        )

        if i == 0:
            literal_specs.append(Stripped(f"{literal_name} {enum_name} = iota"))
        else:
            literal_specs.append(literal_name)

    literal_specs_joined = "\n".join(literal_specs)

    if len(literal_specs) == 0:
        return Stripped(f"type {enum_name} int")

    return Stripped(
        f"""\
// Enumerate the model types for faster type switches.
//
// For example, you can use an array of function pointers to
// implement such a switch.
type {enum_name} int
const (
{I}{indent_but_first_line(literal_specs_joined, I)}
)"""
    )


def _generate_descend_body(
    cls: intermediate.ConcreteClass, recurse: bool, receiver: Identifier
) -> Stripped:
    """
    Generate the body of the `descend and descend-once methods.
    """
    blocks = []  # type: List[Stripped]

    generator_for_loop_variables = golang_common.GeneratorForLoopVariables()

    for prop in cls.properties:
        prop_name = golang_naming.private_property_name(prop.name)

        prop_blocks = []  # type: List[Stripped]

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            continue
        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                continue
            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                continue
            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                prop_blocks.append(
                    Stripped(
                        f"""\
abort = action(
{I}{receiver}.{prop_name},
)
if abort {{
{I}return
}}"""
                    )
                )

                if recurse:
                    prop_blocks.append(
                        Stripped(
                            f"""\
abort = {receiver}.{prop_name}.Descend(
{I}action,
)
if abort {{
{I}return
}}"""
                        )
                    )

            else:
                assert_never(type_anno.our_type)

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin, 2023-03-29): We expect only lists of classes "
                f"at the moment, but you specified {type_anno}. "
                f"Please contact the developers if you need this feature."
            )

            loop_var = next(generator_for_loop_variables)

            if not recurse:
                prop_blocks.append(
                    Stripped(
                        f"""\
for _, {loop_var} := range {receiver}.{prop_name} {{
{I}abort = action({loop_var});
{I}if abort {{
{II}return
{I}}}
}}"""
                    )
                )
            else:
                prop_blocks.append(
                    Stripped(
                        f"""\
for _, {loop_var} := range {receiver}.{prop_name} {{
{I}abort = action({loop_var});
{I}if abort {{
{II}return
{I}}}

{I}abort = {loop_var}.Descend(
{II}action,
{I});
{I}if abort {{
{II}return
{I}}}
}}"""
                    )
                )
        else:
            assert_never(type_anno)

        block = Stripped("\n".join(prop_blocks))
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            block = Stripped(
                f"""\
if {receiver}.{prop_name} != nil {{
{I}{indent_but_first_line(block, I)}
}}"""
            )

        blocks.append(block)

    if len(blocks) == 0:
        blocks.append(Stripped("// No descendable properties"))

    blocks.append(Stripped("return"))

    return Stripped("\n\n".join(blocks))


def _generate_descend_once_method(
    cls: intermediate.ConcreteClass, receiver: Identifier
) -> Stripped:
    """Generate the descend-once method for the concrete class ``cls``."""
    body = _generate_descend_body(cls=cls, recurse=False, receiver=receiver)

    struct_name = golang_naming.struct_name(cls.name)

    return Stripped(
        f"""\
// Apply the action on the instances referenced from {receiver}.
//
// If any of the actions returns abort `true`, the descent is immediately
// stopped,  and abort `true` is also returned. Otherwise, return abort `false`.
//
// We do not recurse into the referenced instances.
//
// The action is not applied on {receiver}.
func ({receiver} *{struct_name}) DescendOnce(
{I}action func(IClass) bool,
) (abort bool) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_descend_method(
    cls: intermediate.ConcreteClass, receiver: Identifier
) -> Stripped:
    """Generate the recursive ``descend`` method for the concrete class ``cls``."""
    body = _generate_descend_body(cls=cls, recurse=True, receiver=receiver)

    struct_name = golang_naming.struct_name(cls.name)

    return Stripped(
        f"""\
// Apply the action recursively on the instances referenced from {receiver}.
//
// If any of the actions returns abort `true`, the descent is immediately
// stopped,  and abort `true` is also returned. Otherwise, return abort `false`.
//
// The action is not applied on {receiver}.
func ({receiver} *{struct_name}) Descend(
{I}action func(IClass) bool,
) (abort bool) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


@require(lambda cls: not cls.is_implementation_specific)
@require(lambda cls: not cls.constructor.is_implementation_specific)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_constructor(
    cls: intermediate.ClassUnion,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the constructor function for the given concrete class ``cls``."""
    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    # NOTE (mristin, 2023-03-31):
    # Golang is peculiar, so we do not transpile the in-lined statements, but simply set
    # the required properties to argument values and optional values to default argument
    # values, if any were specified.
    #
    # It is not clear at this moment whether the SDK clients would really appreciate
    # the added complexity of the code, so we wait till we get some feedback on it.

    blocks = []  # type: List[str]

    struct_specs = []  # type: List[Stripped]

    for arg in cls.constructor.arguments:
        if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        arg_name = golang_naming.argument_name(arg.name)
        private_prop_name = golang_naming.private_property_name(arg.name)

        struct_specs.append(Stripped(f"{private_prop_name}: {arg_name}"))

        if arg.default is not None:
            return None, Error(
                arg.default.parsed.node,
                f"(mristin, 2023-03-31): "
                f"The argument {arg.name!r} is a required argument, "
                f"but the default value is also specified. At the moment when we "
                f"wrote the generator, we did not know how this use case should "
                f"work out in Golang. Please contact the developers and "
                f"discuss how this should be implemented.",
            )

    for arg in cls.constructor.arguments:
        if not isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        if arg.default is None:
            continue

        private_prop_name = golang_naming.private_property_name(arg.name)

        # NOTE (mristin, 2023-03-31):
        # We have to be careful: an optional property will be generated with a pointer
        # type.

        if isinstance(arg.default, intermediate.DefaultPrimitive):
            if arg.default.value is None:
                struct_specs.append(Stripped(f"{private_prop_name}: nil"))
            elif isinstance(arg.default.value, bool):
                literal = golang_common.boolean_literal(arg.default.value)
                struct_specs.append(
                    Stripped(
                        f"""\
{private_prop_name}: aascommon.NewBool(
{I}{literal}
)"""
                    )
                )
            elif isinstance(arg.default.value, int):
                literal = Stripped(str(arg.default.value))
                struct_specs.append(
                    Stripped(
                        f"""\
{private_prop_name}: aascommon.NewInt64(
{I}{literal}
)"""
                    )
                )
            elif isinstance(arg.default.value, float):
                literal = golang_common.float_literal(arg.default.value)
                struct_specs.append(
                    Stripped(
                        f"""\
{private_prop_name}: aascommon.NewFloat64(
{I}{literal}
)"""
                    )
                )
            elif isinstance(arg.default.value, str):
                literal = golang_common.string_literal(arg.default.value)
                struct_specs.append(
                    Stripped(
                        f"""\
{private_prop_name}: aascommon.NewString(
{I}{literal}
)"""
                    )
                )
            else:
                assert_never(arg.default.value)
        elif isinstance(arg.default, intermediate.DefaultEnumerationLiteral):
            literal = golang_naming.enum_literal_name(
                arg.default.enumeration.name, arg.default.literal.name
            )
            enum_name = golang_naming.enum_name(arg.default.enumeration.name)
            struct_specs.append(
                Stripped(
                    f"""\
{private_prop_name}: (*{enum_name})(
{I}aascommon.NewInt(
{II}int({literal})
{I})
)"""
                )
            )
        else:
            assert_never(arg.default)

    struct_name = golang_naming.struct_name(cls.name)

    if len(struct_specs) == 0:
        blocks.append(
            Stripped(
                f"""\
// Intentionally empty.
return &{struct_name}{{}}"""
            )
        )
    else:
        struct_specs_joined = "\n".join(
            f"{struct_spec}," for struct_spec in struct_specs
        )

        blocks.append(
            Stripped(
                f"""\
return &{struct_name}{{
{I}{indent_but_first_line(struct_specs_joined, I)}
}}"""
            )
        )

    body = "\n\n".join(blocks)

    arg_codes = []  # type: List[str]
    for arg in cls.constructor.arguments:
        if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation):
            continue

        arg_type = golang_common.generate_type(type_annotation=arg.type_annotation)
        arg_name = golang_naming.argument_name(arg.name)

        arg_codes.append(Stripped(f"{arg_name} {arg_type}"))

    function_name = golang_naming.function_name(Identifier(f"new_{cls.name}"))

    comment = f"""\
// Create a new instance of {struct_name} with
// the given properties."""

    if len(arg_codes) == 0:
        return (
            Stripped(
                f"""\
{comment}
func {function_name}() *{struct_name} {{
{I}{indent_but_first_line(body, I)}
}}"""
            ),
            None,
        )

    arg_codes_joined = "\n".join(f"{arg_code}," for arg_code in arg_codes)
    return (
        Stripped(
            f"""\
{comment}
func {function_name}(
{I}{indent_but_first_line(arg_codes_joined, I)}
) *{struct_name} {{
{I}{indent_but_first_line(body, I)}
}}"""
        ),
        None,
    )


@require(lambda cls, prop: id(prop) in cls.property_id_set)
@require(lambda prop: prop.description is not None)
def _generate_comment_for_property(
    cls: intermediate.ClassUnion,
    prop: intermediate.Property,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given property."""
    # NOTE (mristin, 2022-10-29):
    # We need to write a double assertion for mypy.
    assert prop.description is not None

    # fmt: off
    comment, errors = (
        golang_description
        .generate_comment_for_summary_remarks_constraints(
            description=prop.description,
            context=golang_description.Context(
                package=golang_common.TYPES_PACKAGE,
                cls_or_enum=cls,
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_interface(
    cls: intermediate.ClassUnion,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the Golang code of the interface corresponding to ``cls``."""
    # Code blocks separated by double newlines and indented once
    blocks = []  # type: List[Stripped]

    # region Inheritance

    if len(cls.inheritances) > 0:
        for inheritance in cls.inheritances:
            inheritance_name = golang_naming.interface_name(inheritance.name)
            blocks.append(inheritance_name)
    else:
        blocks.append(Stripped("IClass"))

    # endregion

    # region Getters and Setters

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        prop_type = golang_common.generate_type(type_annotation=prop.type_annotation)

        prop_comment = None  # type: Optional[Stripped]
        if prop.description is not None:
            (
                prop_comment,
                prop_comment_errors,
            ) = _generate_comment_for_property(cls=cls, prop=prop)

            if prop_comment_errors is not None:
                return None, Error(
                    prop.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for the property {prop.name!r}",
                    prop_comment_errors,
                )

            assert prop_comment is not None

        getter_name = golang_naming.getter_name(prop.name)
        if prop_comment:
            blocks.append(
                Stripped(
                    f"""\
{prop_comment}
{getter_name}() {prop_type};"""
                )
            )
        else:
            blocks.append(Stripped(f"func {getter_name}() {prop_type};"))

        setter_name = golang_naming.setter_name(prop.name)
        blocks.append(
            Stripped(
                f"""\
{setter_name}(
{I}value {prop_type},
);"""
            )
        )

    # endregion

    # region Methods

    for method in cls.methods:
        method_blocks = []  # type: List[Stripped]

        if method.description is not None:
            (
                method_comment,
                method_comment_errors,
            ) = golang_description.generate_comment_for_signature(
                method.description,
                context=golang_description.Context(
                    package=golang_common.TYPES_PACKAGE, cls_or_enum=cls
                ),
            )

            if method_comment_errors is not None:
                return None, Error(
                    method.description.parsed.node,
                    f"Failed to generate the documentation comment "
                    f"for method {method.name!r}",
                    method_comment_errors,
                )

            assert method_comment is not None

            method_blocks.append(method_comment)

        # fmt: off
        returns = (
            f" {golang_common.generate_type(type_annotation=method.returns)}"
            if method.returns is not None else ""
        )
        # fmt: on

        arg_codes = []  # type: List[Stripped]
        for arg in method.arguments:
            arg_type = golang_common.generate_type(type_annotation=arg.type_annotation)
            arg_name = golang_naming.argument_name(arg.name)
            arg_codes.append(Stripped(f"{arg_name} {arg_type}"))

        method_name = golang_naming.method_name(method.name)

        if len(arg_codes) == 0:
            method_blocks.append(Stripped(f"{method_name}(){returns};"))
        else:
            arg_codes_joined = "\n".join(f"{arg_code}," for arg_code in arg_codes)
            method_blocks.append(
                Stripped(
                    f"""\
{method_name}(
{I}{indent_but_first_line(arg_codes_joined, I)}
){returns};"""
                )
            )

        blocks.append(Stripped("\n".join(method_blocks)))

    comment = None  # type: Optional[Stripped]
    if cls.description is not None:
        # fmt: off
        comment, comment_errors = (
            golang_description
            .generate_comment_for_summary_remarks_constraints(
                cls.description,
                context=golang_description.Context(
                    package=golang_common.TYPES_PACKAGE,
                    cls_or_enum=cls
                )
            )
        )
        # fmt: on

        if comment_errors is not None:
            return None, Error(
                cls.description.parsed.node,
                "Failed to generate the documentation comment",
                comment_errors,
            )

        assert comment is not None

    if len(blocks) == 0:
        blocks = [Stripped("// Intentionally empty.")]

    writer = io.StringIO()
    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    interface_name = golang_naming.interface_name(cls.name)

    blocks_joined = "\n\n".join(blocks)

    writer.write(
        f"""\
type {interface_name} interface {{
{I}{indent_but_first_line(blocks_joined, I)}
}}"""
    )

    return Stripped(writer.getvalue()), None


def _generate_is_interface(
    cls: intermediate.ClassUnion, symbol_table: intermediate.SymbolTable
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the function to check the model type against ``cls``."""
    interface_name = golang_naming.interface_name(cls.name)
    function_name = golang_naming.function_name(Identifier(f"is_{cls.name}"))

    if len(cls.concrete_descendants) == 0:
        if isinstance(cls, intermediate.AbstractClass):
            return None, Error(
                cls.parsed.node,
                f"Unexpected abstract class with no concrete descendants: {cls.name!r}",
            )

        model_type_literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )
        body = Stripped(f"ok = that.ModelType() == {model_type_literal}")
    else:
        case_statements = []  # type: List[Stripped]

        for another_cls in symbol_table.concrete_classes:
            model_type_literal = golang_naming.enum_literal_name(
                Identifier("Model_type"), another_cls.name
            )

            if id(another_cls) in cls.concrete_descendant_id_set or another_cls is cls:
                case_statements.append(
                    Stripped(
                        f"""\
case {model_type_literal}:
{I}ok = true"""
                    )
                )

        case_statements_joined = "\n".join(case_statements)
        body = Stripped(
            f"""\
switch that.ModelType() {{
{case_statements_joined}
}}"""
        )

    return (
        Stripped(
            f"""\
// Check whether the instance corresponds to [aastypes.{interface_name}]
// based on its run-time model type.
//
// The implementation uses a switch statements which is
// most probably compiled as an efficient jump table by the compiler.
func {function_name}(
{I}that IClass,
) (ok bool) {{
{I}{indent_but_first_line(body, I)}
{I}return
}}"""
        ),
        None,
    )


@require(lambda cls: not cls.is_implementation_specific)
def _generate_struct(cls: intermediate.ConcreteClass) -> Stripped:
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        prop_type = golang_common.generate_type(type_annotation=prop.type_annotation)
        private_prop_name = golang_naming.private_property_name(prop.name)

        blocks.append(Stripped(f"{private_prop_name} {prop_type}"))

    body = "\n".join(blocks)

    struct_name = golang_naming.struct_name(cls.name)
    interface_name = golang_naming.interface_name(cls.name)

    return Stripped(
        f"""\
// Implements {interface_name}.
type {struct_name} struct {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


@require(lambda cls: not cls.is_implementation_specific)
@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_struct_methods(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[Stripped]], Optional[Error]]:
    """Generate the Golang methods for the struct corresponding to ``cls``."""
    methods = []  # type: List[Stripped]

    receiver = golang_naming.receiver_name(cls)

    # region Setter and getter implementations

    struct_name = golang_naming.struct_name(cls.name)

    for prop in cls.properties:
        prop_type = golang_common.generate_type(prop.type_annotation)
        private_prop_name = golang_naming.private_property_name(prop.name)

        getter_name = golang_naming.getter_name(prop.name)

        methods.append(
            Stripped(
                f"""\
func ({receiver} *{struct_name}) {getter_name}(
) {prop_type} {{
{I}return {receiver}.{private_prop_name}
}}"""
            )
        )

        setter_name = golang_naming.setter_name(prop.name)

        methods.append(
            Stripped(
                f"""\
func ({receiver} *{struct_name}) {setter_name}(
{I}value {prop_type},
) {{
{I}{receiver}.{private_prop_name} = value
}}"""
            )
        )

    # endregion

    # region Getter of Model Type

    model_type_literal = golang_naming.enum_literal_name(
        enumeration_name=Identifier("Model_type"), literal_name=cls.name
    )

    model_type_enum = golang_naming.enum_name(Identifier("Model_type"))
    model_type_getter = golang_naming.getter_name(Identifier("model_type"))

    methods.append(
        Stripped(
            f"""\
func ({receiver} *{struct_name}) {model_type_getter}(
) {model_type_enum} {{
{I}return {model_type_literal}
}}"""
        )
    )

    # endregion

    # region Methods

    errors = []  # type: List[Error]

    for method in cls.methods:
        if isinstance(method, intermediate.ImplementationSpecificMethod):
            implementation_key = specific_implementations.ImplementationKey(
                f"Types/{method.specified_for.name}/{method.name}.go"
            )

            implementation = spec_impls.get(implementation_key, None)

            if implementation is None:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"The implementation is missing for "
                        f"the implementation-specific method: {implementation_key}",
                    )
                )
                continue

            # fmt: off
            methods.append(
                Stripped(
                    implementation
                    .replace("_RECEIVER_", receiver)
                    .replace("_STRUCT_NAME_", struct_name)
                )
            )
            # fmt: on
        else:
            errors.append(
                Error(
                    cls.parsed.node,
                    "(mristin, 2023-03-31) "
                    "At the moment, we do not transpile the method body and "
                    "its contracts. We want to finish the meta-model for the V3, "
                    "fix de/serialization and generate SDKs for a couple of languages "
                    "before taking on this rather hard task.",
                )
            )

    methods.append(_generate_descend_once_method(cls=cls, receiver=receiver))

    methods.append(_generate_descend_method(cls=cls, receiver=receiver))

    # endregion

    # region Constructor

    if cls.constructor.is_implementation_specific:
        implementation_key = specific_implementations.ImplementationKey(
            f"Types/{cls.name}/{cls.name}.go"
        )
        implementation = spec_impls.get(implementation_key, None)

        if implementation is None:
            errors.append(
                Error(
                    cls.parsed.node,
                    f"The implementation of the implementation-specific constructor "
                    f"is missing: {implementation_key}",
                )
            )
        else:
            methods.append(implementation)
    else:
        constructor_block, error = _generate_constructor(cls=cls)

        if error is not None:
            errors.append(error)
        else:
            assert constructor_block is not None

            methods.append(constructor_block)

    # endregion

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the methods for the class {cls.name}",
            errors,
        )

    return methods, None


def _generate_comment_for_meta_model(
    description: intermediate.DescriptionOfMetaModel,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the docstring for the given meta-model."""
    # fmt: off
    comment, errors = (
        golang_description
        .generate_comment_for_summary_remarks_constraints(
            description=description,
            context=golang_description.Context(
                package=golang_common.TYPES_PACKAGE, cls_or_enum=None
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: VerifiedIntermediateSymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Golang code of the structures based on the symbol table.

    The ``aas_module`` indicates the fully-qualified name of the base module.
    """
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    if symbol_table.meta_model.description is not None:
        # fmt: off
        comment, comment_errors = (
            _generate_comment_for_meta_model(
                description=symbol_table.meta_model.description
            )
        )
        # fmt: on

        if comment_errors is not None:
            errors.extend(comment_errors)
        else:
            assert comment is not None
            blocks.append(
                Stripped(
                    f"""\
// Package types provides the data structures corresponding to the meta-model.
//
{comment}
package types"""
                )
            )

    model_type_getter = golang_naming.getter_name(Identifier("model_type"))
    model_type_enum = golang_naming.enum_name(Identifier("Model_type"))

    blocks.extend(
        [
            golang_common.WARNING,
            _generate_definition_for_model_type(symbol_table=symbol_table),
            Stripped(
                f"""\
// Represent the most general interface of an AAS model.
type IClass interface {{
{I}// Return the concrete model type at run-time.
{I}//
{I}// Use the model type if you want to switch on the concrete model type
{I}// in efficient manner, as the compiler will most probably implement
{I}// the switch in form of a jump table.
{I}{model_type_getter}() {model_type_enum}

{I}// Apply the action on the instances referenced from this instance.
{I}//
{I}// If any of the actions returns abort `true`, the descent is immediately
{I}// stopped,  and abort `true` is also returned. Otherwise, return abort `false`.
{I}//
{I}// We do not recurse into the referenced instances.
{I}//
{I}// The action is not applied on this instance.
{I}DescendOnce(action func(IClass) bool) (abort bool)

{I}// Apply the action recursively on the instances referenced from this instance.
{I}//
{I}// If any of the actions returns abort `true`, the descent is immediately
{I}// stopped,  and abort `true` is also returned. Otherwise, return abort `false`.
{I}//
{I}// The action is not applied on this instance.
{I}Descend(action func(IClass) bool) (abort bool)
}}"""
            ),
        ]
    )

    for our_type in symbol_table.our_types:
        if isinstance(our_type, intermediate.Enumeration):
            block, error = _generate_enum(enum=our_type)
            if error is not None:
                errors.append(error)
                continue
            else:
                assert block is not None
                blocks.append(block)
                blocks.append(_literals_of_enum(enum=our_type))

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2023-03-31):
            # We do not generate the constrained primitives as types. We only
            # consider them in the verification.
            continue

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if our_type.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"Types/{our_type.name}.go"
                )

                block = spec_impls.get(implementation_key, None)
                if block is None:
                    errors.append(
                        Error(
                            our_type.parsed.node,
                            f"The implementation is missing "
                            f"for the implementation-specific "
                            f"class: {implementation_key}",
                        )
                    )
                else:
                    blocks.append(block)
            else:
                block, error = _generate_interface(cls=our_type)
                if error is not None:
                    errors.append(error)
                else:
                    assert block is not None
                    blocks.append(block)

                block, error = _generate_is_interface(
                    cls=our_type, symbol_table=symbol_table
                )
                if error is not None:
                    errors.append(error)
                else:
                    assert block is not None
                    blocks.append(block)

                if isinstance(our_type, intermediate.ConcreteClass):
                    block = _generate_struct(cls=our_type)
                    blocks.append(block)

                    methods, error = _generate_struct_methods(
                        cls=our_type, spec_impls=spec_impls
                    )
                    if error is not None:
                        errors.append(error)
                    else:
                        assert methods is not None
                        blocks.extend(methods)
        else:
            assert_never(our_type)

    if len(errors) > 0:
        return None, errors

    blocks.append(golang_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


# endregion



"""Generate Golang code for enhancing model classes."""

import io
from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations
from aas_core_codegen.common import (
    Error,
    Stripped,
    Identifier,
    assert_never,
    indent_but_first_line,
)
from aas_core_codegen.golang import common as golang_common, naming as golang_naming
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


@require(lambda cls, method: id(method) in cls.method_id_set)
def _generate_method_delegation(
    cls: intermediate.ConcreteClass, method: intermediate.Method
) -> Stripped:
    """Generate the delegated method to ``instance``."""
    returns = (
        golang_common.generate_type(
            method.returns, types_package=Identifier("aastypes")
        )
        if method.returns is not None
        else None
    )

    returns_suffix = "" if returns is None else f" {returns}"

    arg_types_names = [
        (
            golang_common.generate_type(
                arg.type_annotation, types_package=Identifier("aastypes")
            ),
            golang_naming.argument_name(arg.name),
        )
        for arg in method.arguments
    ]

    method_name = golang_naming.method_name(method.name)

    return_prefix = "return " if method.returns is not None else ""

    struct_name = golang_naming.private_struct_name(Identifier(f"enhanced_{cls.name}"))
    receiver = golang_naming.receiver_name(cls)

    if len(method.arguments) == 0:
        return Stripped(
            f"""\
func ({receiver} *{struct_name}[E]) {method_name}(){returns_suffix} {{
{I}{return_prefix}{receiver}.instance.{method_name}()
}}"""
        )

    arguments_definition = ",".join(
        f"{arg_name} {arg_type}," for arg_type, arg_name in arg_types_names
    )

    arguments_delegation = "\n".join(f"{arg_name}," for _, arg_name in arg_types_names)

    return Stripped(
        f"""\
func ({receiver} *{struct_name}[E]) {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
){returns_suffix} {{
{I}{return_prefix}{receiver}.instance.{method_name}(
{II}{indent_but_first_line(arguments_delegation, II)}
{I})
}}"""
    )


def _generate_enhanced_struct_and_its_methods(
    cls: intermediate.ConcreteClass,
) -> List[Stripped]:
    """Generate the `enhanced*` struct and its methods as delegation to ``instance``."""
    enhanced_struct_name = golang_naming.private_struct_name(
        Identifier(f"Enhanced_{cls.name}")
    )
    interface_name = golang_naming.interface_name(cls.name)

    model_type_enum = golang_naming.enum_name(Identifier("Model_type"))
    model_type_getter = golang_naming.getter_name(Identifier("model_type"))

    # NOTE (mristin, 2023-05-12):
    # Add an "e" prefix to signal that it is enhanced.
    receiver = golang_naming.receiver_name(cls, prefix="enhanced_")

    result = [
        Stripped(
            f"""\
type {enhanced_struct_name}[E any] struct {{
{I}instance aastypes.{interface_name}
{I}enhancement E
}}"""
        ),
        Stripped(
            f"""\
func ({receiver} *{enhanced_struct_name}[E]) {model_type_getter}(
) aastypes.{model_type_enum} {{
{I}return {receiver}.instance.{model_type_getter}()
}}"""
        ),
        Stripped(
            f"""\
func ({receiver} *{enhanced_struct_name}[E]) DescendOnce(
{I}action func(aastypes.IClass)bool,
) bool {{
{I}return {receiver}.instance.DescendOnce(action)
}}"""
        ),
        Stripped(
            f"""\
func ({receiver} *{enhanced_struct_name}[E]) Descend(
{I}action func(aastypes.IClass) bool,
) bool {{
{I}return {receiver}.instance.Descend(action)
}}"""
        ),
    ]

    for prop in cls.properties:
        prop_type = golang_common.generate_type(
            type_annotation=prop.type_annotation, types_package=Identifier("aastypes")
        )

        getter_name = golang_naming.getter_name(prop.name)
        result.append(
            Stripped(
                f"""\
func ({receiver} *{enhanced_struct_name}[E]) {getter_name}(
) {prop_type} {{
{I}return {receiver}.instance.{getter_name}()
}}"""
            )
        )

        setter_name = golang_naming.setter_name(prop.name)
        result.append(
            Stripped(
                f"""\
func ({receiver} *{enhanced_struct_name}[E]) {setter_name}(
{I}value {prop_type},
) {{
{I}{receiver}.instance.{setter_name}(value)
}}"""
            )
        )

    for method in cls.methods:
        result.append(_generate_method_delegation(cls=cls, method=method))

    result.append(
        Stripped(
            f"""\
func ({receiver} *{enhanced_struct_name}[E]) getEnhancement(
) E {{
{I}return {receiver}.enhancement
}}"""
        )
    )

    result.append(
        Stripped(
            f"""\
func ({receiver} *{enhanced_struct_name}[E]) setEnhancement(
{I}value E,
) {{
{I}{receiver}.enhancement = value
}}"""
        )
    )

    return result


def _generate_wrap_for_cls(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the wrapping function for the concrete class."""
    interface_name = golang_naming.interface_name(cls.name)
    function_name = golang_naming.private_function_name(Identifier(f"wrap_{cls.name}"))

    enhanced_struct_name = golang_naming.private_struct_name(
        Identifier(f"enhanced_{cls.name}")
    )

    recurse_blocks = []  # type: List[Stripped]
    for prop in cls.properties:
        recurse_block: Stripped

        type_anno = intermediate.beneath_optional(prop.type_annotation)

        prop_getter_name = golang_naming.getter_name(prop.name)
        prop_setter_name = golang_naming.setter_name(prop.name)

        prop_var = golang_naming.variable_name(Identifier(f"the_{prop.name}"))

        if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
            # Nothing to recurse into.
            continue

        elif isinstance(type_anno, intermediate.OurTypeAnnotation):
            if isinstance(type_anno.our_type, intermediate.Enumeration):
                # Nothing to recurse into.
                continue

            elif isinstance(type_anno.our_type, intermediate.ConstrainedPrimitive):
                # Nothing to recurse into.
                continue

            elif isinstance(
                type_anno.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ):
                prop_interface_name = golang_naming.interface_name(
                    type_anno.our_type.name
                )
                recurse_block = Stripped(
                    f"""\
that.{prop_setter_name}(
{I}Wrap[E](
{II}{prop_var},
{II}factory,
{I}).(aastypes.{prop_interface_name}),
)"""
                )

            else:
                assert_never(type_anno.our_type)

        elif isinstance(type_anno, intermediate.ListTypeAnnotation):
            assert isinstance(
                type_anno.items, intermediate.OurTypeAnnotation
            ) and isinstance(
                type_anno.items.our_type,
                (intermediate.AbstractClass, intermediate.ConcreteClass),
            ), (
                f"NOTE (mristin, 2023-03-29): We expect only lists of classes "
                f"at the moment, but you specified {type_anno}. "
                f"Please contact the developers if you need this feature."
            )

            items_interface_name = golang_naming.interface_name(
                type_anno.items.our_type.name
            )

            recurse_block = Stripped(
                f"""\
for i, v := range {prop_var} {{
{I}// Update in-situ
{I}{prop_var}[i] = Wrap[E](
{II}v,
{II}factory,
{I}).(aastypes.{items_interface_name})
}}"""
            )

        else:
            assert_never(type_anno)

        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            recurse_block = Stripped(
                f"""\
{prop_var} := that.{prop_getter_name}()
if {prop_var} != nil {{
{I}{indent_but_first_line(recurse_block, I)}
}}"""
            )
        else:
            recurse_block = Stripped(
                f"""\
{prop_var} := that.{prop_getter_name}()
{recurse_block}"""
            )

        recurse_blocks.append(recurse_block)

    blocks = [
        Stripped(
            """\
// We assume that we already checked whether `that` has been enhanced
// in the caller."""
        ),
        Stripped(
            f"""\
enh, shouldEnhance := factory(that)
if shouldEnhance {{
{I}result = &{enhanced_struct_name}[E]{{
{II}instance: that,
{II}enhancement: enh,
{I}}}
}} else {{
{I}result = that
}}"""
        ),
    ]

    if len(recurse_blocks) > 0:
        blocks.extend(recurse_blocks)

    blocks.append(Stripped("return"))

    body = "\n\n".join(blocks)

    return Stripped(
        f"""\
func {function_name}[E any](
{I}that aastypes.{interface_name},
{I}factory func(aastypes.IClass) (E, bool),
) (result aastypes.{interface_name}) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_wrap(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the `Wrap` function."""
    model_type_getter = golang_naming.getter_name(Identifier("model_type"))
    blocks = [
        Stripped(
            f"""\
_, ok := that.(enhanced[E])
if ok {{
{I}panic(
{II}fmt.Sprintf(
{III}"An instance of %T has been already wrapped: %v",
{III}that, that,
{II}),
{I})
}}"""
        )
    ]  # type: List[Stripped]

    case_blocks = []  # type: List[Stripped]
    for cls in symbol_table.concrete_classes:
        literal = golang_naming.enum_literal_name(
            enumeration_name=Identifier("Model_type"), literal_name=cls.name
        )

        interface_name = golang_naming.interface_name(cls.name)
        wrap_function = golang_naming.private_function_name(
            Identifier(f"wrap_{cls.name}")
        )

        case_blocks.append(
            Stripped(
                f"""\
case aastypes.{literal}:
{I}result = {wrap_function}[E](
{II}that.(aastypes.{interface_name}),
{II}factory,
{I})"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}panic(
{II}fmt.Sprintf(
{III}"Unexpected model type: %v",
{III}that.{model_type_getter}(),
{II}),
{I})"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)
    blocks.append(
        Stripped(
            f"""\
switch that.{model_type_getter}() {{
{case_blocks_joined}
}}"""
        )
    )

    blocks.append(Stripped("return"))

    body = "\n\n".join(blocks)

    return Stripped(
        f"""\
// Wrap `that` instance recursively with the enhancement produced by the `factory`.
//
// The factory returns the enhancement, and a boolean "should-enhance". If
// the "should-enhance" is false, `that` instance is not enhance, and we simply
// return it. However, we will still continue to enhance the instances referenced
// by `that` instance recursively.
//
// If `that` instance has been already wrapped, panic.
func Wrap[E any](
{I}that aastypes.IClass,
{I}factory func(aastypes.IClass) (E, bool),
) (result aastypes.IClass) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    repo_url: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the Golang code for enhancing model classes with custom wraps."""
    errors = []  # type: List[Error]

    aastypes_url_literal = golang_common.string_literal(f"{repo_url}/types")

    blocks = [
        Stripped(
            """\
// Package enhancing allows for enhancement of model instances with your custom data.
package enhancing"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}aastypes {aastypes_url_literal}
)"""
        ),
        Stripped(
            f"""\
type enhanced[E any] interface {{
{I}// Get the enhancement from the enhanced instance.
{I}getEnhancement() E

{I}// Set the enhancement of the enhanced instance.
{I}setEnhancement(E)
}}"""
        ),
    ]  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_key = specific_implementations.ImplementationKey(
                f"Enhancing/{cls.name}.go"
            )

            implementation = spec_impls.get(implementation_key, None)
            if implementation is None:
                errors.append(
                    Error(
                        cls.parsed.node,
                        f"The enhancing snippet is missing "
                        f"for the implementation-specific "
                        f"class {cls.name}: {implementation_key}",
                    )
                )
                continue
        else:
            blocks.extend(_generate_enhanced_struct_and_its_methods(cls=cls))
            blocks.append(_generate_wrap_for_cls(cls=cls))

    blocks.append(_generate_wrap(symbol_table=symbol_table))
    blocks.extend(
        [
            Stripped(
                f"""\
// Retrieve the enhancement from `that` instance.
//
// Return the enhancement, or `ok` false, if `that` instance has not been
// enhanced.
func Unwrap[E any](that aastypes.IClass) (enhancement E, ok bool) {{
{I}var enh enhanced[E]
{I}enh, ok = that.(enhanced[E])
{I}if !ok {{
{II}return
{I}}}
{I}enhancement = enh.getEnhancement()
{I}return
}}"""
            ),
            Stripped(
                f"""\
// Retrieve the enhancement from `that` instance.
//
// If `that` instance has not been enhanced yet, panic.
func MustUnwrap[E any](that aastypes.IClass) (enhancement E) {{
{I}var ok bool
{I}enhancement, ok = Unwrap[E](that)
{I}if !ok {{
{II}panic(
{III}fmt.Sprintf(
{IIII}"An instance of %T has not been wrapped: %v",
{IIII}that, that,
{III}),
{II})
{I}}}
{I}return
}}"""
            ),
            golang_common.WARNING,
        ]
    )

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None



"""Generate the code to test XML de/serialization of concrete classes."""

import io
from typing import List

from icontract import ensure

from aas_core_codegen import intermediate, naming
from aas_core_codegen.common import Stripped, Identifier, indent_but_first_line
from aas_core_codegen.python import common as python_common, naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


def _generate_test_case(symbol_table: intermediate.SymbolTable) -> Stripped:
    test_methods = []  # type: List[Stripped]

    for cls in symbol_table.concrete_classes:
        xml_class_name = naming.xml_class_name(cls.name)

        test_method_name = python_naming.method_name(Identifier(f"test_{cls.name}"))

        from_iterparse = python_naming.function_name(
            Identifier(f"{cls.name}_from_iterparse")
        )

        from_stream = python_naming.function_name(Identifier(f"{cls.name}_from_stream"))

        from_file = python_naming.function_name(Identifier(f"{cls.name}_from_file"))

        from_str = python_naming.function_name(Identifier(f"{cls.name}_from_str"))

        test_methods.append(
            Stripped(
                f"""\
def {test_method_name}(self) -> None:
{I}for path in sorted(
{II}(
{III}tests.common.TEST_DATA_DIR
{III}/ "Xml"
{III}/ "Expected"
{III}/ {xml_class_name!r}
{II}).glob("**/*.xml")
{I}):
{II}text = path.read_text(encoding="utf-8")
{II}et_concrete = ET.fromstring(text)
{II}tests.common_xmlization.remove_redundant_whitespace(et_concrete)

{II}# region From iterparse
{II}iterator = ET.iterparse(source=io.StringIO(text), events=["start", "end"])
{II}got_from_iterparse = (
{III}aas_xmlization.{from_iterparse}(iterator)
{II})

{II}et_from_iterparse = ET.fromstring(aas_xmlization.to_str(got_from_iterparse))
{II}tests.common_xmlization.remove_redundant_whitespace(et_from_iterparse)
{II}tests.common_xmlization.assert_elements_equal(et_concrete, et_from_iterparse)
{II}# endregion

{II}# region From stream
{II}got_from_stream = (
{III}aas_xmlization
{III}.{from_stream}(
{IIII}io.StringIO(text)
{III})
{II})
{II}et_from_stream = ET.fromstring(aas_xmlization.to_str(got_from_stream))
{II}tests.common_xmlization.remove_redundant_whitespace(et_from_stream)
{II}tests.common_xmlization.assert_elements_equal(et_concrete, et_from_stream)
{II}# endregion

{II}# region From file
{II}with tempfile.TemporaryDirectory() as tmp_dir:
{III}path = pathlib.Path(tmp_dir) / "something.xml"
{III}path.write_text(text, encoding="utf-8")

{III}got_from_file = (
{IIII}aas_xmlization
{IIII}.{from_file}(path)
{III})
{II}et_from_file = ET.fromstring(aas_xmlization.to_str(got_from_file))
{II}tests.common_xmlization.remove_redundant_whitespace(et_from_file)
{II}tests.common_xmlization.assert_elements_equal(et_concrete, et_from_file)
{II}# endregion

{II}# region From string
{II}got_from_str = (
{III}aas_xmlization
{III}.{from_str}(text)
{II})
{II}et_from_str = ET.fromstring(aas_xmlization.to_str(got_from_str))
{II}tests.common_xmlization.remove_redundant_whitespace(et_from_str)
{II}tests.common_xmlization.assert_elements_equal(et_concrete, et_from_str)
{II}# endregion"""
            ),
        )

    body = (
        "\n\n".join(test_methods)
        if len(test_methods) > 0
        else """\
# There are no concrete classes.
pass"""
    )

    return Stripped(
        f"""\
class TestRoundTrips(unittest.TestCase):
{I}{indent_but_first_line(body, I)}"""
    )


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(
    symbol_table: intermediate.SymbolTable,
    qualified_module_name: python_common.QualifiedModuleName,
) -> str:
    """
    Generate the code to test XML de/serialization of concrete classes.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '''\
"""Test XML de/serialization of concrete classes."""'''
        ),
        python_common.WARNING,
        Stripped(
            """\
# pylint: disable=missing-docstring"""
        ),
        Stripped(
            """\
import io
import pathlib
import tempfile
import unittest
import xml.etree.ElementTree as ET"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.xmlization as aas_xmlization"""
        ),
        Stripped(
            """\
import tests.common_xmlization"""
        ),
        _generate_test_case(symbol_table=symbol_table),
        Stripped(
            f"""\
if __name__ == "__main__":
{I}unittest.main()"""
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()
